"""Python utilities for the Forthon fortran wrapper.
"""
"""
But flies an eagle flight, bold, and forthon, Leaving no tract behind.
"""
# import all of the neccesary packages
import sys
import __main__

# --- Only numpy is now supported.
from numpy import *
def gettypecode(x):
    return x.dtype.char

import re
import os
import copy
import warnings
import cPickle
try:
    from PyPDB import PW, PR
except ImportError:
    pass
try:
    import inspect
except ImportError:
    pass
# --- Add line completion capability
try:
    import readline
except ImportError:
    pass
else:
    import rlcompleter
    readline.parse_and_bind("tab: complete")

##############################################################################
# --- Functions needed for object pickling. These should be moved to C.
def forthonobject_constructor(typename, arg=None):
    if isinstance(arg, str):
        mname = arg
    else:
        # --- In old versions, second arg was None or a dict.
        # --- In those cases, use main as the module.
        mname = "__main__"
    m = __import__(mname)
    typecreator = getattr(m, typename)
    if callable(typecreator):
        obj = typecreator()
        # --- For old pickle files, a dict will still be passed in, relying on
        # --- this rather than setstate.
        if isinstance(arg, dict):
            obj.setdict(arg)
        return obj
    else:
        # --- When typecreator is not callable, this means that it is a top
        # --- level package. In this case, just return it since is should
        # --- be restored elsewhere.
        return typecreator
def pickle_forthonobject(o):
    if o.getfobject() == 0:
        # --- For top level Forthon objects (which are package objects
        # --- as opposed to derived type objects) only save the typename.
        # --- This assumes that the package will be written out directly
        #  --- elsewhere.
        return (forthonobject_constructor, (o.gettypename(), o.__module__))
    else:
        # --- The dictionary from getdict will be passed into the __setstate__
        # --- method upon unpickling.
        return (forthonobject_constructor, (o.gettypename(), o.__module__),
                o.getdict())

# --- The following routines deal with multiple packages. The ones setting
# --- up or changing the allocation of groups will be called from fortran.

# --- The currently active package defaults to none. Stick this in the
# --- main dictionary. This is somewhat kludgy but I think this is the only
# --- way. The value is put in main so it can be saved in a restart dump
# --- and have the code restarted with the correct package active.
# --- Also, since it is put directly in main, it does not appear in the
# --- list of globals in warp.py and so is not include in
# --- initial_global_dict_keys which would prevent it from being written
# --- to the dump file.
__main__.__dict__["currpkg"] = ' '
def getcurrpkg():
    return __main__.__dict__["currpkg"]
def setcurrpkg(pkg):
    __main__.__dict__["currpkg"] = pkg

_pkg_dict = {}
_pkg_list = []
def registerpackage(pkg, name):
    """
    Registers a package so it can be accessed using various global routines.
    """

    # --- For each package, which has its own type, the pickling functions
    # --- must be registered for that type. Note that this is not needed
    # --- for packages that are class instances.
    if IsForthonType(pkg):
        import copy_reg
        copy_reg.pickle(type(pkg), pickle_forthonobject, forthonobject_constructor)
    else:
        assert isinstance(pkg, PackageBase),\
            "Only instances of classes inheritting from PackageBase can be registered as a package"

    _pkg_dict[name] = pkg
    _pkg_list.append(name)

def package(name=None):
    """
    Sets the default package - sets global currpkg variable.
    If name is not given, then returns a copy of the list of all registered
    packages.
    """
    if name is None:
        return copy.deepcopy(_pkg_list)
    setcurrpkg(name)
    _pkg_list.remove(name)
    _pkg_list.insert(0, name)

def packageobject(name):
    """
    Returns the package object, rather than the package name
    """
    return _pkg_dict[name]

def gallot(group='*', iverbose=0):
    """
    Allocates all dynamic arrays in the specified group.
    If the group is not given or is '*', then all groups are allocated.
    When optional argument iverbose is true, the variables allocated and their new size is printed.
    """
    for pkg in _pkg_dict.itervalues():
        r = pkg.gallot(group, iverbose)
        if r and (group != '*'):
            return
    if group != '*':
        raise NameError("No such group")

def gchange(group='*', iverbose=0):
    """
    Changes the allocation of all dynamic arrays in the specified group if it is needed.
    If the group is not given or is '*', then all groups are changed.
    When optional argument iverbose is true, the variables changed and their new size is printed.
    """
    for pkg in _pkg_dict.itervalues():
        r = pkg.gchange(group, iverbose)
        if r and (group != '*'):
            return
    if group != '*':
        raise NameError("No such group")

def gfree(group='*'):
    """
    Frees the allocated memory of all dynamic arrays in the specified group.
    If the group is not given or is '*', then all groups are freed.
    """
    for pkg in _pkg_dict.itervalues():
        r = pkg.gfree(group)
        if r and (group != '*'):
            return
    if group != '*':
        raise NameError("No such group")

def gsetdims(group='*'):
    """
    Sets the size in the python database of all dynamic arrays in the specified group.
    If the group is not given or is '*', then it is done for all groups.
    """
    for pkg in _pkg_dict.itervalues():
        r = pkg.gsetdims(group)
        if r and (group != '*'):
            return
    if group != '*':
        raise NameError("No such group")

def forceassign(name, v):
    """
    Forces the assignment to an array, resizing the array is necessary.
    """
    for pkg in _pkg_dict.itervalues():
        pkg.forceassign(name, v)

def listvar(name):
    """
    Lists information about a variable, given the name as a string.
    """
    for pkg in _pkg_dict.itervalues():
        r = pkg.listvar(name)
        if r is not None:
            return r
    raise NameError

def deprefix():
    """
    For each variable in the package, a python object is created which has the same name and
    same value. For arrays, the new objects points to the same memory location.
    """
    pkglist = package()
    pkglist.reverse()
    for pname in pkglist:
        packageobject(pname).deprefix()

def reprefix():
    """
    For each variable in the main dictionary, if there is a package variable with the same name
    it is assigned to that value. For arrays, the data is copied.
    """
    for pkg in _pkg_dict.itervalues():
        pkg.reprefix()

def totmembytes():
    """
    Prints the total amount of memory dynamically allocated for all groups in all packages.
    """
    tot = 0.
    for pkg in _pkg_dict.itervalues():
        tot = tot + pkg.totmembytes()
    return tot

def IsForthonType(v):
    t = repr(type(v))
    if re.search("Forthon", t):
        return 1
    else:
        return 0


# --- Create a base class that can be used as a package object.
# --- This includes all of the necessary methods, any of which
# --- can be overwritten in the inheritting class.
class PackageBase(object):

    def generate(self):
        raise NotImplementedError

    def step(self):
        raise NotImplementedError

    def finish(self):
        raise NotImplementedError

    def gallot(self, group='*', iverbose=0):
        return 0

    def gchange(self, group='*', iverbose=0):
        return 0

    def gfree(self, group='*'):
        return 0

    def gsetdims(self, group='*'):
        return 0

    def forceassign(self, name, v):
        pass

    def listvar(self, name):
        return name

    def deprefix(self):
        pass

    def reprefix(self):
        pass

    def totmembytes(self):
        return getobjectsize(self)


# --- Some platforms have a different value of .true. in fortran.
# --- Note that this is not necessarily correct. The most robust
# --- solution would be to get the value directly from a fortran
# --- routine. As of now though, there are no fortran routines
# --- built into Forthon, only those from the users code.
if sys.platform in ['sn960510']:
    true = -1
    false = 0
else:
    true = 1
    false = 0

# --- Converts an array of characters into a string.
def arraytostr(a, strip=true):
    a = array(a)
    if len(shape(a)) == 1:
        result = a[0]
        if sys.hexversion >= 0x03000000:
            # --- This is needed for Python3, where string arrays give
            # --- numpy.bytes_ objects. It needs to be converted to unicode.
            if isinstance(result, bytes_):
                result = result.decode()
        if strip:
            result = result.strip()
    elif len(shape(a)) == 2:
        result = []
        for i in xrange(shape(a)[1]):
            result.append(arraytostr(a[:, i]))
    return result

# --- Allows int operation on arrrays

builtinint = int

def aint(x):
    if isinstance(x, ndarray):
        return x.astype('l')
    else:
        return builtinint(x)

if sys.hexversion < 0x03000000:
    # --- This is needed for legacy code
    int = aint

# --- Return the nearest integer
def nint(x):
    if isinstance(x, ndarray):
        return where(greater(x, 0), aint(x + 0.5), -aint(abs(x) + 0.5))
    else:
        if x >= 0:
            return int(x + 0.5)
        else:
            return -int(abs(x) + 0.5)

def fones(shape, *args):
    """T
    his is a replacement for the array creation routine, ones, which
    creates arrays which have the proper ordering for fortran. When arrays
    created with this are passed to a fortran subroutine, no copies are needed to
    get the data into the proper order for fortran. It takes the same arguments
    as ones, except for 'order' which is set to F.
    """
    return ones(shape, order='F', *args)

def fzeros(shape, *args):
    """
    This is a replacement for the array creation routine, zeros, which
    creates arrays which have the proper ordering for fortran. When arrays
    created with this are passed to a fortran subroutine, no copies are needed to
    get the data into the proper order for fortran. It takes the same arguments
    as zeros, except for 'order' which is set to F.
    """
    return zeros(shape, order='F', *args)

def doc(f, printit=1):
    """
    Prints out the documentation of the subroutine or variable.
    The name of package variables must be given in quotes, as a string.
    All objects can be passed in directly.
    """
    fname = None
    # --- The for loop only gives the code something to break out of. There's
    # --- probably a better way of doing this.
    for i in range(1):
        if isinstance(f, str):
            # --- Check if it is a Forthon variable
            try:
                d = listvar(f)
                break
            except NameError:
                pass
            # --- Check if it is a module name
            try:
                m = __import__(f)
                d = m.__doc__
                if d is None:
                    try:
                        d = m.__dict__[f + 'doc']()
                    except KeyError:
                        pass
                if d is None:
                    d = ''
                fname = determineoriginatingfile(m)
                break
            except ImportError:
                pass
            # --- Try to get the actual value of the object
            try:
                v = __main__.__dict__[f]
                fname = determineoriginatingfile(v)
                d = v.__doc__
                break
            except KeyError:
                d = "Name not found"
            except AttributeError:
                d = "No documentation found"
        else:
            fname = determineoriginatingfile(f)
            # --- Check if it has a doc string
            try:
                d = f.__doc__
            except AttributeError:
                d = None
    if d is None:
        d = "No documentation found"
    if fname is not None:
        result = 'From file ' + fname + '\n'
    else:
        result = ''
    result += d
    if printit:
        print result
    else:
        return result

def determineoriginatingfile(o):
    """
    Attempts to determine the name of the file where the given object was
    defined.  The input can be an object or a string. If it is a string, it
    tries to find the object in __main__ or looks for a module with that
    name. If it can't find anything, it returns None. If an object is passed in,
    introspection is used to find the file name. Note there are cases
    where the file name can not be determined - None is returned.
    """
    import types
    if isinstance(o, str):
        # --- First, deal with strings
        try:
            # --- Look in __main__
            o = __main__.__dict__[o]
        except KeyError:
            # --- If not there, try to find a module with that name.
            # --- Note: this could return a different module than the original
            # --- if module names are redundant and/or the sys.path has changed.
            try:
                o = __import__(o)
                return determineoriginatingfile(o)
            except ImportError:
                # --- If that fails, just return None
                return None
    # --- Now deal with the various types.
    # --- Note that this is done recursively in some cases to reduce redundant
    # --- coding.
    # --- For all other types, either the information is not available,
    # --- or it doesn't make sense.
    if isinstance(o, types.ModuleType):
        try:
            return o.__file__
        except AttributeError:
            return '%s (statically linked into python)'%o.__name__
    if isinstance(o, (types.MethodType, types.UnboundMethodType)):
        return determineoriginatingfile(o.im_class)
    if isinstance(o, (types.FunctionType, types.LambdaType)):
        return determineoriginatingfile(o.func_code)
    if isinstance(o, (types.BuiltinFunctionType, types.BuiltinMethodType)):
        try:
            m = __import__(o.__module__)
        except AttributeError:
            return None
        if m is not None:
            return determineoriginatingfile(m)
        else:
            return None
    if isinstance(o, types.CodeType):
        return o.co_filename
    if isinstance(o, types.InstanceType):
        return determineoriginatingfile(o.__class__)
    if isinstance(o, (types.ClassType, types.TypeType)):
        return determineoriginatingfile(__import__(o.__module__))

# --- Get size of an object, recursively including anything inside of it.
def oldgetobjectsize(pkg, grp='', recursive=1):
    """
    Gets the total size of a package or dictionary.
      - pkg: Either a Forthon object, dictionary, or a class instance
      - grp='': For a Forthon object, only include the variables in the specified
                group
      - recursive=1: When true, include the size of sub objects.
    """

    # --- Keep track of objects already accounted for.
    # --- The call level is noted so that at the end, at call level zero,
    # --- the list of already accounted for objects can be deleted.
    try:
        if id(pkg) in getobjectsize.grouplist:
            return 0
        getobjectsize.grouplist.append(id(pkg))
        getobjectsize.calllevel += 1
    except AttributeError:
        getobjectsize.grouplist = []
        getobjectsize.calllevel = 0

    # --- Return sizes of shallow objects
    if isinstance(pkg, (int, float)):
        result = 1
    elif isinstance(pkg, ndarray):
        result = product(array(shape(pkg)))
    else:
        result = 0

    # --- Get the list of variables to check. Note that the grp option only
    # --- affects Forthon objects.
    if IsForthonType(pkg):
        ll = pkg.varlist(grp)
    elif isinstance(pkg, dict):
        ll = pkg.iterkeys()
    elif isinstance(pkg, (list, tuple)):
        ll = pkg
    else:
        try:
            ll = pkg.__dict__.iterkeys()
        except AttributeError:
            ll = []

    if not recursive and getobjectsize.calllevel > 0:
        ll = []

    # --- Now, add up the sizes.
    for v in ll:
        if IsForthonType(pkg):
            # --- This is needed so unallocated arrays will only return None
            vv = pkg.getpyobject(v)
        elif isinstance(pkg, dict):
            vv = pkg[v]
        elif isinstance(pkg, (list, tuple)):
            vv = v
        else:
            vv = getattr(pkg, v)
        result = result + getobjectsize(vv, '', recursive=recursive)

    # --- Do some clean up or accounting before exiting.
    if getobjectsize.calllevel == 0:
        del getobjectsize.grouplist
        del getobjectsize.calllevel
    else:
        getobjectsize.calllevel -= 1

    # --- Return the result
    return result

# --- Get size of an object, recursively including anything inside of it.
# --- New improved version, though should be tested more
def getobjectsize(pkg, grp='', recursive=1, grouplist=None):
    """
    Gets the total size of a package or dictionary.
      - pkg: Either a Forthon object, dictionary, or a class instance
      - grp='': For a Forthon object, only include the variables in the specified
                group
      - recursive=1: When true, include the size of sub objects.
    """

    if grouplist is None:
        grouplist = []

    # --- Keep track of objects already accounted for.
    if id(pkg) in grouplist:
        # --- Even the the item has already been counted, add the
        # --- approximate size of the reference
        return 1
    grouplist.append(id(pkg))

    # --- Return sizes of shallow objects
    if isinstance(pkg, (int, float, bool)):
        return 1
    elif isinstance(pkg, ndarray):
        return product(array(shape(pkg)))

    # --- The object itself gets count of 1
    result = 1

    # --- Get the list of variables to check. Note that the grp option only
    # --- affects Forthon objects.
    import collections
    if IsForthonType(pkg):
        ll = pkg.varlist(grp)
    elif isinstance(pkg, dict):
        ll = pkg.iterkeys()
    elif isinstance(pkg, (list, tuple)):
        ll = pkg
    else:
        try:
            ll = pkg.__dict__.iterkeys()
        except (AttributeError, NameError):
            ll = []

    if not recursive and len(grouplist) > 1:
        ll = []

    # --- Now, add up the sizes.
    for v in ll:
        if IsForthonType(pkg):
            # --- This is needed so unallocated arrays will only return None
            vv = pkg.getpyobject(v)
        elif isinstance(pkg, dict):
            result += 1  # --- Add one for the key
            vv = pkg[v]
        elif isinstance(pkg, (ndarray, collections.Sequence)):
            vv = v
        else:
            try:
                vv = getattr(pkg, v)
            except AttributeError:
                vv = 0
        result = result + getobjectsize(vv, '', recursive=recursive, grouplist=grouplist)

    # --- Return the result
    return result

# --- Keep the old name around
getgroupsize = getobjectsize

def getgroupsizes(pkg, minsize=1, sortby='sizes'):
    """
    Get the sizes of groups in the specified package.
     - pkg: package to list
     - minsize=1: only groups with size greater than the given value are printed
     - sortby='sizes': When 'sizes', sort by sizes, when 'names', sort by names,
                       otherwise unsorted.
    """
    groups = {}
    varlist = pkg.varlist()
    for n in varlist:
        group = pkg.getgroup(n)
        v = pkg.getpyobject(n)
        groups[group] = groups.get(group, 0) + size(v)

    if sortby == 'sizes':
        ii = argsort(groups.values())
        keysunsorted = groups.keys()
        keys = []
        for i in ii:
            keys.append(keysunsorted[i])

    elif sortby == 'names':
        keys = groups.keys()
        keys.sort()

    else:
        keys = groups.keys()

    for k in keys:
        v = groups[k]
        if v > minsize:
            print k, v, '(words)'

    print "Total size of allocated arrays", pkg.totmembytes()

# --- Print out all variables in a group
def printgroup(pkg, group='', maxelements=10, sumarrays=0):
    """
    Print out all variables in a group or with an attribute
      - pkg: package name or class instance (where group is ignored)
      - group: group name
      - maxelements=10: only up to this many elements of arrays are printed
      - sumarrays=0: when true, prints the total sum of arrays rather than the
                     first several elements
    """
    if isinstance(pkg, str):
        pkg = __main__.__dict__[pkg]
    try:
        vlist = pkg.varlist(group)
    except AttributeError:
        vlist = pkg.__dict__.iterkeys()
    if not vlist:
        print "Unknown group name " + group
        return
    for vname in vlist:
        try:
            v = pkg.getpyobject(vname)
        except AttributeError:
            v = pkg.__dict__[vname]
        if v is None:
            print vname + ' is not allocated'
        elif not isinstance(v, ndarray):
            print vname + ' = ' + str(v)
        else:
            if gettypecode(v) == 'c':
                print vname + ' = "' + str(arraytostr(v)) + '"'
            elif sumarrays:
                sumv = sum(reshape(v, tuple([product(array(v.shape))])))
                print 'sum(' + vname + ') = ' + str(sumv)
            elif size(v) <= maxelements:
                print vname + ' = ' + str(v)
            else:
                if rank(v) == 1:
                    print vname + ' = ' + str(v[:maxelements])[:-1] + " ..."
                else:
                    if shape(v)[0] <= maxelements:
                        if rank(v) == 2:
                            print vname + ' = [' + str(v[:,0]) + "] ..."
                        elif rank(v) == 3:
                            print vname + ' = [[' + str(v[:,0,0]) + "]] ..."
                        elif rank(v) == 4:
                            print vname + ' = [[[' + str(v[:,0,0,0]) + "]]] ..."
                        elif rank(v) == 5:
                            print vname + ' = [[[[' + str(v[:,0,0,0,0]) + "]]]] ..."
                        elif rank(v) == 6:
                            print vname + ' = [[[[[' + str(v[:,0,0,0,0,0]) + "]]]]] ..."
                    else:
                        if rank(v) == 2:
                            print vname + ' = [' + str(v[:maxelements,0])[:-1] + " ..."
                        elif rank(v) == 3:
                            print vname + ' = [[' + str(v[:maxelements,0,0])[:-1] + " ..."
                        elif rank(v) == 4:
                            print vname + ' = [[[' + str(v[:maxelements,0,0,0])[:-1] + " ..."
                        elif rank(v) == 5:
                            print vname + ' = [[[[' + str(v[:maxelements,0,0,0,0])[:-1] + " ..."
                        elif rank(v) == 6:
                            print vname + ' = [[[[[' + str(v[:maxelements,0,0,0,0,0])[:-1] + " ..."

##############################################################################
##############################################################################
def pydumpforthonobject(ff, attr, objname, obj, varsuffix, writtenvars, serial, verbose):
    """Loops over the variables in the Forthon object and writes each out if requested
    All variables are written directly to the datawriter (assuming that it can handle
    arbitrary objects, such as Forthon derived types).
    """
    if verbose:
        print "object " + objname + " being written"
    # --- Get variables in this package which have attribute attr.
    vlist = []
    for a in attr:
        if isinstance(a, str):
            vlist = vlist + obj.varlist(a)
    # --- Loop over list of variables
    for vname in vlist:
        # --- Get the object if it is available
        v = obj.getpyobject(vname)
        if v is None:
            # --- If not available, skip it (dynamic arrays may be unallocated for example)
            if verbose:
                print "variable " + vname + varsuffix + " skipped since it is not available"
            continue
        # --- If serial flag is set, check if it has the parallel attribute
        if serial:
            a = obj.getvarattr(vname)
            if re.search('parallel', a):
                if verbose:
                    print "variable " + vname + varsuffix + " skipped since it is a parallel variable"
                continue
        # --- Check if variable with same name has already been written out.
        # --- This only matters when the variable is being written out as
        # --- a plane python variable.
        if '@' not in varsuffix:
            if vname in writtenvars:
                if verbose:
                    print "variable " + objname + "." + vname + " skipped since other variable would have same name in the file"
                continue
            writtenvars.append(vname)
        # --- If this point is reached, then variable is written out to file
        if verbose:
            print "writing " + objname + "." + vname + " as " + vname + varsuffix
        # --- Write it out to the file
        ff.write(vname + varsuffix, v)

##############################################################################
# Python version of the dump routine. This uses the varlist command to
# list of all of the variables in each package which have the
# attribute attr (and actually attr could be a group name too). It then
# checks on the state of the python object, making sure that unallocated
# arrays are not written out.  Finally, the variable is written out to the
# file with the name in the format vame@pkg.  Additionally, python
# variables can be written to the file by passing in a list of the names
# through vars. The '@' sign is used between the package name and the
# variable name so that no python variable names can be clobbered ('@'
# is not a valid character in python names). The 'ff.write' command is
# used, allowing names with an '@' in them. The writing of python variables
# is put into a 'try' command since some variables cannot be written to
# a dump file.
def pydump(fname=None, attr=["dump"], vars=[], serial=0, ff=None, varsuffix=None,
           verbose=false, hdf=0, datawriter=None):
    """
    Dump data into a pdb file
      - fname: dump file name
      - attr=["dump"]: attribute or list of attributes of variables to dump
           Any items that are not strings are skipped. To write no variables,
           use attr=None.
      - vars=[]: list of python variables to dump
      - serial=0: switch between parallel and serial versions
      - ff=None: Allows passing in of a file object so that pydump can be called
           multiple times to pass data into the same file. Note that
           the file must be explicitly closed by the user.
           ff can be an instance of any class that conforms to the API of PW.
      - varsuffix=None: Suffix to add to the variable names. If none is specified,
           the suffix '@pkg' is used, where pkg is the package name that the
           variable is in. Note that if varsuffix is specified, the simulation
           cannot be restarted from the dump file.
      - verbose=false: When true, prints out the names of the variables as they are
           written to the dump file
      - hdf=0: (obsolete) this argument is ignored
      - datawriter=PW.PW: datawriter is the data writer class to use. This can be any
                          class that conforms to the API of PW.PW from the PyPDB package.
    """
    assert fname is not None or ff is not None,\
        "Either a filename must be specified or a data writer instance"
    if hdf:
        warnings.warn("the hdf argument is no longer used and is ignored")
    # --- Open the file if the file object was not passed in.
    # --- If the file object was passed in, then don't close it.
    if ff is None:
        if datawriter is None:
            # --- PyPDB is the default data format.
            try:
                datawriter = PW.PW
            except NameError:
                pass

        # --- Try to open the file using the datawriter.
        if datawriter is not None:
            try:
                ff = datawriter(fname)
            except IOError:
                raise
            except:
                pass

        assert ff is not None, "Dump file cannot be created, the datawriter cannot open the file or is unspecified"
        closefile = 1
    else:
        closefile = 0

    # --- Make sure the file has a file_type. Older versions of the pdb
    # --- wrapper did not define a file type.
    try:
        ff.file_type
    except AttributeError:
        ff.file_type = 'unknown'

    if verbose:
        print "Data will be written using %s format"%ff.file_type

    # --- Convert attr into a list if needed
    if not isinstance(attr, list):
        attr = [attr]

    # --- Loop through all of the packages (getting pkg object).
    # --- When varsuffix is specified, the list of variables already written
    # --- is created. This solves two problems. It gives proper precedence to
    # --- variables of the same name in different packages. It also fixes
    # --- an obscure bug in the pdb package - writing two different arrays with
    # --- the same name causes a problem and the pdb file header is not
    # --- properly written. The pdb code should really be fixed.
    pkgsuffix = varsuffix
    packagelist = package()
    writtenvars = []
    for pname in packagelist:
        pkg = packageobject(pname)
        if isinstance(pkg, PackageBase):
            continue
        if varsuffix is None:
            pkgsuffix = '@' + pname
        pydumpforthonobject(ff, attr, pname, pkg, pkgsuffix, writtenvars, serial, verbose)
        # --- Make sure that pname does not appear in vars
        try:
            vars.remove(pname)
        except ValueError:
            pass

    # --- Now, write out the python variables (that can be written out).
    # --- If supplied, the varsuffix is append to the names here too.
    import types
    if varsuffix is None:
        varsuffix = ''
    for vname in vars:
        # --- Skip python variables that would overwrite fortran variables.
        if len(writtenvars) > 0:
            if vname in writtenvars:
                if verbose:
                    print "variable " + vname + " skipped since other variable would have same name in the file"
                continue
        # --- Get the value of the variable.
        vval = __main__.__dict__[vname]
        # --- Write out the source of functions. Note that the source of functions
        # --- typed in interactively is not retrieveable - inspect.getsource
        # --- returns an IOError.
        # --- Callable class instances do not need to be treated specially here.
        if isinstance(vval, types.FunctionType):
            source = None
            try:
                # --- Check if the source had been saved as an attribute of itself.
                # --- This allows functions to be saved that would otherwise
                # --- not be because inspect.getsource can't find them.
                source = vval.__source__
            except AttributeError:
                pass
            if source is None:
                try:
                    source = inspect.getsource(vval)
                except (IOError, NameError, TypeError):
                    pass
            if source is not None:
                if verbose:
                    print "writing python function " + vname + " as " + vname + varsuffix + '@function'
                # --- Clean up any indentation in case the function was defined in
                # --- an indented block of code
                while source[0] == ' ':
                    source = source[1:]
                # --- Now write it out
                ff.write(vname + varsuffix + '@function', source)
                # --- Save the source of a function as an attribute of itself to make
                # --- retreival easier the next time.
                setattr(vval, '__source__', source)
            else:
                if verbose:
                    print "could not write python function " + vname
            continue
        # --- Zero length arrays cannot by written out.
        if isinstance(vval, ndarray) and product(array(shape(vval))) == 0:
            continue
        # --- Try writing as normal variable.
        try:
            if verbose:
                print "writing python variable " + vname + " as " + vname + varsuffix
            ff.write(vname + varsuffix, vval)
            continue
        except:
            pass
        # --- If that didn't work, try writing as a pickled object
        # --- This is only needed for the old pdb wrapper. The new one
        # --- automatically pickles things as needed.
        if ff.file_type == 'unknown':
            try:
                if verbose:
                    print "writing python variable " + vname + " as " + vname + varsuffix + '@pickle'
                ff.write(vname + varsuffix + '@pickle', cPickle.dumps(vval, -1))
                docontinue = 1
            except (cPickle.PicklingError, TypeError):
                pass
            if docontinue:
                continue
        # --- All attempts failed so write warning message
        if verbose:
            print "cannot write python variable " + vname
    if closefile:
        ff.close()

#############################################################################
# Python version of the restore routine. It restores all of the variables
# in the pdb file.
# An '@' in the name distinguishes between the two. The 'ff.__getattr__' is
# used so that variables with an '@' in the name can be read. The reading
# in of python variables is put in a 'try' command to make it idiot proof.
# More fancy foot work is done to get new variables read in into the
# global dictionary.
def pyrestore(filename=None, fname=None, verbose=0, skip=[], ff=None,
              varsuffix=None, ls=0, lreturnfobjdict=0, lreturnff=0,
              datareader=None, main=None):
    """
    Restores all of the variables in the specified file.
      - filename: file to read in from (assumes PDB format)
      - verbose=0: When true, prints out the names of variables which are read in
      - skip=[]: list of variables to skip
      - ff=None: Allows passing in of a file object so that pydump can be called
           multiple times to pass data into the same file. Note that
           the file must be explicitly closed by the user.
           ff can be an instance of any class that conforms to the API of PR.
      - varsuffix: when set, all variables read in will be given the suffix
                   Note that fortran variables are then read into python vars
      - ls=0: when true, prints a list of the variables in the file
              when 1 prints as tuple
              when 2 prints in a column
      - datareader=PR.PR: data reader object, can be any class that conforms to the
                          API of the PR.PR class from PyPDB
      - main=__main__: main object that Forthon objects are restored into
                       Used when the Forthon package is not "import *" into main.
    Note that it will automatically detect whether the file is PDB or HDF.
    """
    # --- fname is the old input argument name
    if filename is None:
        filename = fname
    assert filename is not None or ff is not None,\
        "Either a filename must be specified or a data reader instance"
    if ff is None:
        if datareader is None:
            try:
                # --- PyPDB is the default data format.
                datareader = PR.PR
            except NameError:
                pass

        # --- Check if file exists
        assert os.access(filename, os.F_OK), "File %s does not exist"%filename

        # --- Try opening file with either the default PR or user supplied reader
        try:
            ff = datareader(filename)
        except:
            pass

        assert ff is not None, "File %s could not be opened"%filename
        closefile = 1
    else:
        closefile = 0

    if lreturnff:
        closefile = 0

    # --- Make sure the file has a file_type. Older versions of the pdb
    # --- wrapper did not define a file type.
    try:
        ff.file_type
    except AttributeError:
        ff.file_type = 'unknown'

    if verbose:
        print "Data will be read using %s format"%ff.file_type

    # --- Get a list of all of the variables in the file, loop over that list
    vlist = ff.inquire_names()

    # --- Print list of variables
    if ls:
        if ls == 1:
            print vlist
        else:
            for l in vlist:
                print l

    # --- First, sort out the list of variables
    groups = sortrestorevarsbysuffix(vlist, skip)
    fobjdict = {}

    # --- Read in the variables with the standard suffices.

    # --- These would be interpreter variables written to the file
    # --- from python (or other sources). A simple assignment is done and
    # --- the variable in put in the main dictionary.
    if '' in groups:
        plist = groups['']
        del groups['']
        for vname in plist:
            pyname = vname
            if varsuffix is not None:
                pyname = pyname + str(varsuffix)
            try:
                if verbose:
                    print "reading in python variable " + vname
                __main__.__dict__[pyname] = ff.__getattr__(vname)
            except:
                if verbose:
                    print "error with variable " + vname

    # --- These would be interpreter variables written to the file
    # --- as pickled objects. The data is unpickled and the variable
    # --- in put in the main dictionary.
    # --- This is only needed with the old pdb wrapper.
    if ff.file_type == 'unknown':
        if 'pickle' in groups:
            picklelist = groups['pickle']
            del groups['pickle']
            for vname in picklelist:
                pyname = vname
                if varsuffix is not None:
                    pyname = pyname + str(varsuffix)
                try:
                    if verbose:
                        print "reading in pickled variable " + vname
                    __main__.__dict__[pyname] = cPickle.loads(ff.__getattr__(vname + '@pickle'))
                except:
                    if verbose:
                        print "error with variable " + vname

    # --- These would be interpreter variables written to the file
    # --- from Basis. A simple assignment is done and the variable
    # --- in put in the main dictionary.
    if 'global' in groups:
        globallist = groups['global']
        del groups['global']
        for vname in globallist:
            pyname = vname
            if varsuffix is not None:
                pyname = pyname + str(varsuffix)
            try:
                if verbose:
                    print "reading in Basis variable " + vname
                __main__.__dict__[pyname] = ff.__getattr__(vname + '@global')
            except:
                if verbose:
                    print "error with variable " + vname

    # --- User defined Python functions
    import types
    if 'function' in groups:
        functionlist = groups['function']
        del groups['function']
        for vname in functionlist:
            # --- Skip functions which have already been defined in case the user
            # --- has made source updates since the dump was made. But only skip
            # --- if the existing thing is actually a function.
            if vname in __main__.__dict__ and isinstance(__main__.__dict__[vname], types.FunctionType):
                if verbose:
                    print "skipping python function %s since it already is defined"%vname
            else:
                try:
                    if verbose:
                        print "reading in python function" + vname
                    source = ff.__getattr__(vname + '@function')
                    exec(source, __main__.__dict__)
                    # --- Save the source of the function as an attribute of itself
                    # --- so that it can be latter saved in a dump file again.
                    # --- This is needed since for any functions defined here,
                    # --- inspect.getsource cannot get the source.
                    setattr(__main__.__dict__[vname], '__source__', source)
                except:
                    if verbose:
                        print "error with function " + vname

    # --- Ignore variables with suffix @parallel
    if 'parallel' in groups:
        del groups['parallel']

    for gname in groups.iterkeys():
        pyrestoreforthonobject(main, ff, gname, groups[gname], fobjdict, varsuffix,
                               verbose, doarrays=0, main=main)
    for gname in groups.iterkeys():
        pyrestoreforthonobject(main, ff, gname, groups[gname], fobjdict, varsuffix,
                               verbose, doarrays=1, main=main)

    if closefile:
        ff.close()
    resultlist = []
    if lreturnfobjdict:
        resultlist.append(fobjdict)
    if lreturnff:
        resultlist.append(ff)
    if len(resultlist) == 1:
        return resultlist[0]
    elif len(resultlist) > 1:
        return resultlist

def sortrestorevarsbysuffix(vlist, skip):
    # --- Sort the variables, collecting them in groups based on their suffix.
    groups = {}
    for v in vlist:
        if '@' in v:
            i = v.rfind('@')
            vname = v[:i]
            gname = v[i+1:]
        else:
            # --- Otherwise, variable is plain python variable.
            vname = v
            gname = ''

        # --- If variable is in the skip list, then skip
        if (vname in skip or (len(v) > 4 and v[-4]=='@' and v[-3:] + '.' + v[:-4] in skip)):
            continue

        # --- Now add the variable to the appropriate group list.
        groups.setdefault(gname, []).append(vname)

    return groups

# -----------------------------------------------------------------------------
def pyrestoreforthonobject(obj, ff, gname, vlist, fobjdict, varsuffix, verbose, doarrays,
                           gpdbname=None, main=None):
    """
      - obj: Forthon object data is being restored into
      - ff: reference to file being written to
      - gname: name (in python format) of object to read in
      - vlist: list of variables from the file that are part of this object
      - fobjdist: dictionary of objects already read in
      - varsuffix: suffix to apply to all variables (in python)
      - verbose: when true, lists whether each variable is successfully read in
      - doarrays: when true, reads in arrays, otherwise only scalars
      - gpdbname: actual name of object in the data file. If None, extracted
                  from gname.
    """

    if main is None:
        # --- Default is the current top level main object.
        main = __main__

    if obj is None:
        obj = main

    # --- Convert gname in pdb-style name
    gsplit = gname.split('.')
    attrname = gsplit[-1]
    if gpdbname is None:
        gsplit.reverse()
        gpdbname = '@'.join(gsplit)

    # --- Always create a new object except for the top level packages.
    try:
        attrobj = getattr(obj, attrname)
    except AttributeError:
        attrobj = None
    neednew = (attrobj is None)

    if neednew:
        # --- A new variable needs to be created.
        try:
            fobj = ff.__getattr__("FOBJ@" + gpdbname)
        except:
            return
        # --- First, check if the object has already been restored.
        if fobj in fobjdict:
            # --- If so, then point new variable to existing object
            attrobj = fobjdict[fobj]
            setattr(obj, attrname, attrobj)
            # main.__dict__[gname] = main.__dict__[fobjdict[fobj]]
            # exec("%s = %s"%(gname, fobjdict[fobj]), main.__dict__)
            # return ???
        else:
            # --- Otherwise, create a new instance of the appropriate type,
            # --- and add it to the list of objects.
            typename = ff.__getattr__("TYPENAME@" + gpdbname)
            try:
                attrobj = main.__dict__[typename]()
                setattr(obj, attrname, attrobj)
                # main.__dict__[gname] = main.__dict__[typename]()
                # exec("%s = %s()"%(gname, typename), main.__dict__)
            except:
                # --- If it gets here, it might mean that the name no longer exists.
                return
            fobjdict[fobj] = attrobj

    # --- Sort out the list of variables
    groups = sortrestorevarsbysuffix(vlist, [])

    # --- Get "leaf" variables
    if '' in groups:
        leafvars = groups['']
        del groups['']
    else:
        leafvars = []

    # --- Fix the case when the variable ff appears in the main dictionary.
    # --- This messes up the exec commands below
    def doassignment(fullname, val):
        # --- This loops over the attributes, finding the second to the last
        # --- level. setattr is then used to assign the value to the leaf.
        n = fullname.split('.')
        v = main.__dict__[n[0]]
        for a in n[1:-1]:
            v = getattr(v, a)
        setattr(v, n[-1], val)

    # --- Read in leafs.
    for vname in leafvars:
        if vname == 'FOBJ' or vname == 'TYPENAME':
            continue
        fullname = gname + '.' + vname

        # --- Add suffix to name if given.
        # --- varsuffix is wrapped in str in case a nonstring was passed in.
        if varsuffix is not None:
            fullname = vname + str(varsuffix)

        try:
            val = ff.__getattr__(vname + '@' + gpdbname)
            if not isinstance(val, ndarray) and not doarrays:
                # --- Simple assignment is done for scalars, using the exec command
                if verbose:
                    print "reading in " + fullname
                # doassignment(fullname, val)
                setattr(attrobj, vname, val)
            elif isinstance(val, ndarray) and doarrays:
                if verbose:
                    print "reading in " + fullname
                if varsuffix is None:
                    if (len(str(val.dtype)) > 1 and str(val.dtype)[1] == 'S' and
                       getattr(attrobj, vname).shape != val.shape):
                        # --- This is a crude fix for backwards compatibility. The way
                        # --- strings are handled changed, so that they now have an
                        # --- element size > 1. This coding converts old style strings
                        # --- into a single string before doing the setattr. The change
                        # --- affects restart dumps make before July 2008.
                        setattr(attrobj, vname, ''.join(val))
                    else:
                        setattr(attrobj, vname, val)
                else:
                    # --- If varsuffix is specified, then put the variable directly into
                    # --- the main dictionary.
                    # doassignment(fullname, val)
                    setattr(attrobj, vname + str(varsuffix), val)
        except:
            # --- The catches errors in cases where the variable is not an
            # --- actual variable in the package, for example if it had been deleted
            # --- after the dump was originally made.
            print "Warning: There was a problem restoring %s"%(fullname)
            # --- Print out information about exactly what went wrong.
            if verbose:
                sys.excepthook(*sys.exc_info())

    # --- Read in rest of groups.
    for g, v in groups.iteritems():
        pyrestoreforthonobject(getattr(obj, attrname), ff, gname + '.' + g, v, fobjdict, varsuffix, verbose, doarrays,
                               g + '@' + gpdbname, main=main)


# --- create an alias for pyrestore
restore = pyrestore

##############################################################################
##############################################################################
##############################################################################
##############################################################################
##############################################################################
##############################################################################

def Forthondoc():
    print """
package(): sets active package or returns list of all packages
gallot(): allocates all dynamic arrays in a group
gchange(): changes all dynamic arrays in a group if needed
gfree(): free all dynamic arrays in a group
gsetdims(): setups up dynamic arrays sizes in datebase
forceassign(): forces assignment to a dynamic array, resizing if necessary
listvar(): prints information about a variable
deprefix(): creates a python variable for each package variable
reprefix(): copies python variables into packages variables of the same name
totmembytes(): returns total memory allocated for dynamic arrays
PackageBase: Base class for classes that can be registered as a package
arraytostr(): converts an array of chars to a string
int(): converts data to integer
nint(): converts data to nearest integer
fones(): returns multi-dimensional array with fortran ordering
fzeros(): returns multi-dimensional array with fortran ordering
doc(): prints info about variables and functions
printgroup(): prints all variables in the group or with an attribute
pydump(): dumps data into pdb format file
pyrestore(): reads data from pdb format file
restore(): equivalent to pyrestore
"""



"""Routine timer and profiler
These are generic classes which provide timing information about routine
calls and traces.
ForthonProfiler: accumulates function level timings and optionally prints
                 traces on function calls and returns
enablelinetracing: enables line level tracing
disablelinetracing: disables line level tracing
"""
import sys
import time
import linecache


def ForthonTimerdoc():
    import ForthonTimer
    print ForthonTimer.__doc__


#############################################################################
# --- These two classes create a profiler which keeps
# --- track on function executions and timings of them. Instances of the
# --- first class stores the data for each function. The second class does
# --- the setup and contains the profiler.
class ForthonTimings:
    """
    This is a holder for function calls and timing information. Each instance
    holds the timing info for a subroutine and references to the instances for its
    caller and all of its callees.
     - level: depth of routine call from the toplevel
     - name: name of the routine
     - parent: class instance of the caller of this routine
    """

    def __init__(self, level, name, parent=None):
        self.level = level
        self.name = name
        self.parent = parent
        self.time = 0.
        self.ncalls = 0
        self.subtimers = {}
        self.renew()

    def renew(self):
        """
        This restarts the timing clock and increments the number of calls made to
        this function.
        """
        self.starttime = time.clock()
        self.ncalls += 1

    def newtimer(self, name):
        """
        This returns a timer to be used for a callee. If the callee has already be
        called before, return that timer with the time reset. Otherwise create a
        new one and return it.
        """
        if name in self.subtimers:
            self.subtimers[name].renew()
        else:
            self.subtimers[name] = ForthonTimings(self.level+1, name, self)
        return self.subtimers[name]

    def stoptimer(self):
        """
        Stop the timing clock and accumulate the time. This assumes that this is
        called on a function return, so it returns the timer instance of the caller
        (the parent).
        """
        self.endtime = time.clock()
        self.time = self.time + self.endtime - self.starttime
        return self.parent

    def out(self, maxlevel, mintime=0.):
        """
        Prints info about the function and all of its callees, up to the input level.
        """
        if self.level > maxlevel:
            return
        if self.time > mintime:
            print "%2d%s%s %d %f"%(self.level, self.level*'  ', self.name, self.ncalls, self.time)
        for v in self.subtimers.itervalues():
            v.out(maxlevel, mintime)


class ForthonProfiler:
    """
    This class installs a profiler which gathers timing information for all python
    functions and methods called. It keeps a list of ForthonTimings instances
    for top level functions (those called from the same name space where the
    instance is created, or from the interpreter). Each of those contain all of
    the info about their callees.
    Argument:
      - trace=0: when true, a trace of the routines called and returned from will
                 be actively printed
      - tracelevel=None: when specified, trace is only printed in levels less than
                         the given value. The trace option is automatically
                         switched on.
    """

    _ninstances = 0

    def __init__(self, trace=0, tracelevel=None):
        if ForthonProfiler._ninstances > 0:
            raise RuntimeError("Only one instance allowed.")
        ForthonProfiler._ninstances = 1
        self.trace = trace
        self.tracelevel = tracelevel
        if self.tracelevel is not None:
            self.trace = 1
        self.restart()

    def restart(self):
        self.root = None
        self.level = 0
        self.finished = 0
        sys.setprofile(self.profiler)

    def finish(self):
        """
        Turn off the profiling. This also stops the clock of the top level routine.
        """
        if self.finished:
            return
        sys.setprofile(None)
        self.root.stoptimer()
        self.finished = 1

    def profiler(self, frame, event, arg):
        """
        This is the profile routine. It creates the instances of ForthonTimings for each
        routine called and starts and stops the timers.
        """
        # --- Get the name of the routine
        name = frame.f_code.co_name
        # --- If name is not known, this could mean that this was called from
        # --- inside the sys.setprofile routine, or from some other odd place,
        # --- like from fortran. Skip those cases.
        if name == '?':
            return
        # --- Turn the profiler off during these operations (though this is
        # --- probably unecessary since the sys package should already do this).
        sys.setprofile(None)
        # --- Create an instance of the timer for the toplevel if it hasn't
        # --- already been done.
        if self.root is None:
            self.root = ForthonTimings(0, "toplevel")
            self.timer = self.root
        if event == 'return' and self.level > 0:
            self.level = self.level - 1
            self.timer = self.timer.stoptimer()
        if self.trace:
            if self.tracelevel is None or self.tracelevel > self.level:
                # --- The flush is added so that anything that was printed to stdout
                # --- or stderr get outputed now so that the print outs occur
                # --- at the proper time relative to the trace.
                sys.stdout.flush()
                sys.stderr.flush()
                print "%s %s %s"%(self.level*'  ', event, name)
        if event == 'call':
            self.level = self.level + 1
            self.timer = self.timer.newtimer(name)
        # --- Turn the profiler back on
        sys.setprofile(self.profiler)

    def out(self, maxlevel=2, mintime=0.):
        """
        Print out timing info.
         - maxlevel=2: only prints timings up to the given call depth
         - mintime=0.: only prints timings greater than or equal to the given value
        """
        self.finish()
        self.root.out(maxlevel, mintime)


###############################################################################
# --- Thanks to Andrew Dalke
# --- See http://www.dalkescientific.com/writings/diary/archive/2005/04/20/tracing_python_code.html
def traceit(frame, event, arg):
    if event == "line":
        lineno = frame.f_lineno
        try:
            filename = frame.f_globals["__file__"]
        except KeyError:
            filename = ''
        if (filename.endswith(".pyc") or filename.endswith(".pyo")):
            filename = filename[:-1]
        name = frame.f_globals["__name__"]
        line = linecache.getline(filename, lineno)
        print "%s:%s: %s" % (name, lineno, line.rstrip())
    return traceit


def enablelinetracing():
    """
    Enables line by line tracing.
    """
    sys.settrace(traceit)


def disablelinetracing():
    """
    Disables line by line tracing.
    """
    sys.settrace(None)



"""Generates the wrapper for derived types.
"""
import sys
import fvars
import string
import hashlib
from cfinterface import *


class ForthonDerivedType:
    def __init__(self, typelist, pkgname, pkgsuffix, pkgbase, c, f, isz, writemodules, fcompname):
        if not typelist:
            return

        self.pkgname = pkgname
        self.pkgsuffix = pkgsuffix
        self.pkgbase = pkgbase

        self.cfile = open(c, 'a')
        self.ffile = open(f, 'a')
        self.wrapderivedtypes(typelist, pkgname, pkgsuffix, isz, writemodules, fcompname)
        self.cfile.close()
        self.ffile.close()

    transtable = (10*string.ascii_lowercase)[:256]

    def fsub(self, type, prefix, suffix='', dohash=1):
        """
        The fortran standard limits routine names to 31 characters. If the
        routine name is longer than that, this routine takes the first 15
        characters and and creates a hashed string based on the full name to get
        the next 16. This does not guarantee uniqueness, but the nonuniqueness
        should be minute.
        """
        name = type.name + prefix + suffix
        if len(name) < 32 or not dohash:
            return name
        transtable = ForthonDerivedType.transtable
        if sys.hexversion >= 0x03000000:
            hashbytes = hashlib.md5(name.encode()).digest()
            hash = ''.join([transtable[d] for d in hashbytes])
        else:
            hash = hashlib.md5(name).digest().translate(transtable)
        return name[:15] + hash

    def dimisparameter(self, dim):
        # --- Convert fortran variable name into reference from list of variables
        # --- and check if it is a parameter.
        sl = re.split('[ ()/\*\+\-]', dim)
        for ss in sl:
            if re.search('[a-zA-Z]', ss) is not None:
                try:
                    v = self.slist[self.sdict[ss]]
                    if v.parameter:
                        return True
                except KeyError:
                    pass
        return False

    def prefixdimsc(self, dim):
        # --- Convert fortran variable name into reference from list of variables.
        sl = re.split('[ ()/\*\+\-]', dim)
        for ss in sl:
            if re.search('[a-zA-Z]', ss) is not None:
                if ss in self.sdict:
                    dim = re.sub(ss, '*(long *)obj->fscalars[' + repr(self.sdict[ss]) + '].data',
                                 dim, count=1)
                else:
                    raise SyntaxError('%s from dim %s is not declared in a .v file'%(ss, dim))
        return dim.lower()

    # --- Convert variable names in to type elements
    def prefixdimsf(self, dim):
        sl = re.split('[ ()/\*\+\-:,]', dim)
        # --- Loop over the list of items twice. The first time, add in the 'obj%'
        # --- prefix but overwrite the item with '=='. Then go back again and
        # --- overwrite the '==' with the item. This is done in case one name
        # --- has another in it or a item is repeated. Note that this could also
        # --- be done instead by using re syntax in the substring being searched
        # --- for, by forcing complete words to be found, without a leading % sign.
        # --- But this works as is and is fast enough.
        for ss in sl:
            if re.search('[a-zA-Z]', ss) is not None:
                if ss in self.sdict:
                    dim = re.sub(ss, 'fobj__%==', dim, count=1)
        for ss in sl:
            if re.search('[a-zA-Z]', ss) is not None:
                if ss in self.sdict:
                    dim = re.sub('==', ss, dim, count=1)
        # --- Check for any unspecified dimensions and replace it with an element
        # --- from the dims array.
        sl = re.split(',', dim[1:-1])
        for i in range(len(sl)):
            if sl[i] == ':':
                sl[i] = 'dims__(%d)'%(i + 1)
        dim = '(' + ','.join(sl) + ')'
        return dim.lower()

    def getmodulename(self):
        if self.pkgbase is not None:
            return self.pkgbase
        else:
            return self.pkgname + self.pkgsuffix + 'py'

    # --------------------------------------------
    def cw(self, text, noreturn=0):
        if noreturn:
            self.cfile.write(text)
        else:
            self.cfile.write(text + '\n')

    def fw(self, text, noreturn=0):
        i = 0
        while len(text[i:]) > 132 and text[i:].find('&') == -1:
            # --- If the line is too long, then break it up, adding line
            # --- continuation marks in between any variable names.
            # --- This is the same as \W, but also skips %, since PG compilers
            # --- don't seem to like a line continuation mark just before a %.
            ss = re.search('[^a-zA-Z0-9_%]', text[i+130::-1])
            assert ss is not None, "Forthon can't find a place to break up this line:\n" + text
            text = text[:i+130-ss.start()] + '&\n' + text[i+130-ss.start():]
            i += 130 - ss.start() + 1
        if noreturn:
            self.ffile.write(text)
        else:
            self.ffile.write(text + '\n')

    # --- This is the routine that does all of the work for derived types
    def wrapderivedtypes(self, typelist, pkgname, pkgsuffix, isz, writemodules, fcompname):

        for t in typelist:
            self.cw('')
            vlist = t.vlist[:]

            self.cw('static char* %spointee__ = "%s";'%(t.name, t.name))

            # --- Select out all of the scalars and build a dictionary
            # --- The dictionary is used to get number of the variables use as
            # --- dimensions for arrays.
            slist = []
            self.sdict = {}
            i = 0
            temp = vlist[:]
            for v in temp:
                if not v.dims and not v.function:
                    slist.append(v)
                    self.sdict[v.name] = i
                    i = i + 1
                    vlist.remove(v)

            # --- Select out all of the arrays
            alist = []
            i = 0
            temp = vlist[:]
            for v in temp:
                if not v.function:
                    alist.append(v)
                    i = i + 1
                    vlist.remove(v)
            temp = []

            self.slist = slist

            #########################################################################
            # --- Print out the external commands
            self.cw('extern void ' + fname(self.fsub(t, 'passpointers')) + '(char *fobj, ' + 'long *setinitvalues);')
            self.cw('extern void ' + fname(self.fsub(t, 'nullifypointers')) + '(char *fobj);')
            self.cw('extern PyObject *' + fname(self.fsub(t, 'newf')) + '(void);')
            self.cw('extern void ' + fname(self.fsub(t, 'deallocatef')) + '(char *);')
            self.cw('extern void ' + fname(self.fsub(t, 'nullifycobjf')) + '(char *);')

            # --- setpointer and getpointer routine
            # --- Note that setpointer get written out for all derived types -
            # --- for non-dynamic derived types, the setpointer routine does a copy.
            for s in slist:
                if (s.dynamic or s.derivedtype) and not s.parameter:
                    self.cw('extern void ' + fname(self.fsub(t, 'setscalarpointer', s.name)) +
                            '(char *p, char *fobj__, npy_intp *nullit__);')
                if s.dynamic:
                    self.cw('extern void ' + fname(self.fsub(t, 'getscalarpointer', s.name)) +
                            '(ForthonObject **cobj__, char *fobj__, int *createnew__);')
            for a in alist:
                self.cw('extern void ' + fname(self.fsub(t, 'setarraypointer', a.name)) +
                        '(char *p, char *fobj__, npy_intp *dims__);')
                if a.dynamic or re.search('fassign', a.attr):
                    self.cw('extern void ' + fname(self.fsub(t, 'getarraypointer', a.name)) +
                            '(Fortranarray *farray__, char* fobj__);')
            self.cw('')

            # --- setaction and getaction routines
            for s in slist:
                if s.setaction is not None:
                    self.cw('extern void ' + fname(self.fsub(t, 'setaction', s.name)) +
                            '(char *fobj, ' + fvars.ftoc_dict[s.type] + ' *v);')
                if s.getaction is not None:
                    self.cw('extern void ' + fname(self.fsub(t, 'getaction', s.name)) +
                            '(char *fobj);')
            for a in alist:
                if a.setaction is not None:
                    self.cw('extern void ' + fname(self.fsub(t, 'setaction', a.name)) +
                            '(char *fobj, ' + fvars.ftoc_dict[a.type] + ' *v);')
                if a.getaction is not None:
                    self.cw('extern void ' + fname(self.fsub(t, 'getaction', a.name)) +
                            '(char *fobj);')
            self.cw('')

            #########################################################################
            # --- Write declarations of c pointers to fortran variables
            self.cw('void ' + t.name + 'declarevars(ForthonObject *obj) {')

            # --- Scalars
            self.cw('obj->nscalars = ' + repr(len(slist)) + ';')
            if len(slist) > 0:
                self.cw('obj->fscalars = PyMem_Malloc(obj->nscalars*sizeof(Fortranscalar));')
            else:
                self.cw('obj->fscalars = NULL;')
            for i in range(len(slist)):
                s = slist[i]
                if (s.dynamic or s.derivedtype) and not s.parameter:
                    setscalarpointer = '*' + fname(self.fsub(t, 'setscalarpointer', s.name))
                else:
                    setscalarpointer = 'NULL'
                if s.dynamic:
                    getscalarpointer = '*' + fname(self.fsub(t, 'getscalarpointer', s.name))
                else:
                    getscalarpointer = 'NULL'
                if s.setaction is None:
                    setaction = 'NULL'
                else:
                    setaction = '*' + fname(self.fsub(t, 'setaction', s.name))
                if s.getaction is None:
                    getaction = 'NULL'
                else:
                    getaction = '*' + fname(self.fsub(t, 'getaction', s.name))
                self.cw('obj->fscalars[%d].type = NPY_%s;'%(i, fvars.ftop(s.type)))
                self.cw('obj->fscalars[%d].typename = "%s";'%(i, s.type))
                self.cw('obj->fscalars[%d].name = "%s";'%(i, s.name))
                self.cw('obj->fscalars[%d].data = NULL;'%i)
                self.cw('obj->fscalars[%d].group = "%s";'%(i, t.name))
                self.cw('obj->fscalars[%d].attributes = "%s";'%(i, s.attr))
                # The repr is used so newlines get written out as \n.
                self.cw('obj->fscalars[%d].comment = "%s";'%(i, repr(s.comment)[1:-1].replace('"', '\\"')))
                self.cw('obj->fscalars[%d].dynamic = %d;'%(i, s.dynamic))
                self.cw('obj->fscalars[%d].parameter = %d;'%(i, s.parameter))
                self.cw('obj->fscalars[%d].setscalarpointer = %s;'%(i, setscalarpointer))
                self.cw('obj->fscalars[%d].getscalarpointer = %s;'%(i, getscalarpointer))
                self.cw('obj->fscalars[%d].setaction = %s;'%(i, setaction))
                self.cw('obj->fscalars[%d].getaction = %s;'%(i, getaction))

            # --- Arrays
            self.cw('obj->narrays = ' + repr(len(alist)) + ';')
            if len(alist) > 0:
                self.cw('obj->farrays = PyMem_Malloc(obj->narrays*sizeof(Fortranarray));')
            else:
                self.cw('obj->farrays = NULL;')
            for i in range(len(alist)):
                a = alist[i]
                if a.dynamic:
                    setarraypointer = '*' + fname(self.fsub(t, 'setarraypointer', a.name))
                else:
                    setarraypointer = 'NULL'
                if a.dynamic or re.search('fassign', a.attr):
                    getarraypointer = '*' + fname(self.fsub(t, 'getarraypointer', a.name))
                else:
                    getarraypointer = 'NULL'
                if a.setaction is None:
                    setaction = 'NULL'
                else:
                    setaction = '*' + fname(self.fsub(t, 'setaction', a.name))
                if a.getaction is None:
                    getaction = 'NULL'
                else:
                    getaction = '*' + fname(self.fsub(t, 'getaction', a.name))
                if a.data and a.dynamic:
                    initvalue = a.data[1:-1]
                else:
                    initvalue = '0'
                self.cw('obj->farrays[%d].type = NPY_%s;'%(i, fvars.ftop(a.type)))
                self.cw('obj->farrays[%d].dynamic = %d;'%(i, a.dynamic))
                self.cw('obj->farrays[%d].nd = %d;'%(i, len(a.dims)))
                self.cw('obj->farrays[%d].dimensions = (npy_intp*)NULL;'%i)
                self.cw('obj->farrays[%d].name = "%s";'%(i, a.name))
                self.cw('obj->farrays[%d].data.s = (char *)NULL;'%i)
                self.cw('obj->farrays[%d].setarraypointer = %s;'%(i, setarraypointer))
                self.cw('obj->farrays[%d].getarraypointer = %s;'%(i, getarraypointer))
                self.cw('obj->farrays[%d].setaction = %s;'%(i, setaction))
                self.cw('obj->farrays[%d].getaction = %s;'%(i, getaction))
                self.cw('obj->farrays[%d].initvalue = %s;'%(i, initvalue))
                self.cw('obj->farrays[%d].pya = NULL;'%i)
                self.cw('obj->farrays[%d].group = "%s";'%(i, a.group))
                self.cw('obj->farrays[%d].attributes = "%s";'%(i, a.attr))
                self.cw('obj->farrays[%d].comment = "%s";'%(i, repr(a.comment)[1:-1].replace('"', '\\"')))
                self.cw('obj->farrays[%d].dimstring = "%s";'%(i, repr(a.dimstring)[1:-1]))
            self.cw('}')

#     # --- Write out the table of getset routines
#     self.cw('')
#     self.cw('static PyGetSetDef ' + t.name + '_getseters[] = {')
#     for i in range(len(slist)):
#       s = slist[i]
#       if s.type == 'real': gstype = 'double'
#       elif s.type == 'double': gstype = 'double'
#       elif s.type == 'float': gstype = 'float'
#       elif s.type == 'integer': gstype = 'integer'
#       elif s.type == 'complex': gstype = 'cdouble'
#       else:                    gstype = 'derivedtype'
#       self.cw('{"' + s.name + '", (getter)Forthon_getscalar' + gstype +
#                            ', (setter)Forthon_setscalar' + gstype +
#                      ', "%s"'%repr(s.comment)[1:-1].replace('"', '\\"') +
#                           ', (void *)' + repr(i) + '}, ')
#     for i in range(len(alist)):
#       a = alist[i]
#       self.cw('{"' + a.name + '", (getter)Forthon_getarray' +
#                            ', (setter)Forthon_setarray' +
#                      ', "%s"'%repr(a.comment)[1:-1].replace('"', '\\"') +
#                           ', (void *)' + repr(i) + '}, ')
#     self.cw('{"scalardict", (getter)Forthon_getscalardict, ' +
#                           '(setter)Forthon_setscalardict, ' +
#             '"internal scalar dictionary", NULL}, ')
#     self.cw('{"arraydict", (getter)Forthon_getarraydict, ' +
#                          '(setter)Forthon_setarraydict, ' +
#             '"internal array dictionary", NULL}, ')
#     self.cw('{NULL}};')

            #########################################################################
            # --- Write static array initialization routines
            self.cw('void ' + t.name + 'setstaticdims(ForthonObject *obj)')
            self.cw('{')

            i = -1
            for a in alist:
                i = i + 1
                vname = 'obj->farrays[%d]'%i
                if a.dims and not a.dynamic:
                    j = 0
                    for d in a.dims:
                        if d.high == '':
                            continue
                        self.cw('   ' + vname + '.dimensions[' + repr(j) + '] = (npy_intp)((int)',
                                noreturn=1)
                        j = j + 1
                        if re.search('[a-zA-Z]', d.high) is None:
                            self.cw('(' + d.high + ') - ', noreturn=1)
                        else:
                            if not self.dimisparameter(d.high):
                                raise SyntaxError('%s: static dims must be constants or parameters'%a.name)
                            self.cw('(' + self.prefixdimsc(d.high) + ') - ', noreturn=1)
                        if re.search('[a-zA-Z]', d.low) is None:
                            self.cw('(' + d.low + ') + 1);')
                        else:
                            if not self.dimisparameter(d.low):
                                raise SyntaxError('%s: static dims must be constants or parameters'%a.name)
                            self.cw('(' + self.prefixdimsc(d.low) + ') + 1);')
            self.cw('}')

            #########################################################################
            # --- Write routine which sets the dimensions of the dynamic arrays.
            # --- This is done in a seperate routine so it only appears once.
            # --- A routine is written out for each group which has dynamic arrays.
            # --- Then a routine is written which calls all of the individual group
            # --- routines. That is done to reduce the strain on the compiler by
            # --- reducing the size of the routines. (In fact, in one case, with
            # --- everything in one routine, the cc compiler was giving a core dump!)
            # --- Loop over the variables. This assumes that the variables are sorted
            # --- by group.
            self.cw('static void ' + t.name + 'setdims(char *name, ForthonObject *obj, long i)')
            self.cw('{')

            i = -1
            for a in alist:
                i = i + 1
                vname = 'obj->farrays[%d]'%i
                if a.dynamic == 1 or a.dynamic == 2:
                    j = 0
                    self.cw('  if (i == -1 || i == %d) {'%i)
                    # --- create lines of the form dims[1] = high - low + 1
                    for d in a.dims:
                        if d.high == '':
                            continue
                        self.cw('   ' + vname + '.dimensions[' + repr(j) + '] = (npy_intp)((int)',
                                noreturn=1)
                        j = j + 1
                        if re.search('[a-zA-Z]', d.high) is None:
                            self.cw('(' + d.high + ') - ', noreturn=1)
                        else:
                            self.cw('(' + self.prefixdimsc(d.high) + ') - ', noreturn=1)
                        if re.search('[a-zA-Z]', d.low) is None:
                            self.cw('(' + d.low + ') + 1);')
                        else:
                            self.cw('(' + self.prefixdimsc(d.low) + ') + 1);')
                    self.cw('  }')
            self.cw('}')

            #########################################################################
            # --- Routines to deal with garbage collection. This is only needed
            # --- if a derived type contains pointers to objects of its own type.
            garbagecollected = 0
            # --- Check if type contains pointers to its own type
            for s in slist:
                if s.dynamic and s.type == t.name:
                    garbagecollected = 1

            #########################################################################
            #########################################################################
            self.cw('PyObject *' + pkgname + '_' + t.name + 'New(PyObject *self, PyObject *args)')
            self.cw('{')
            self.cw('  return ' + fname(self.fsub(t, 'newf')) + '();')
            self.cw('}')
            #########################################################################
            # --- Write out an empty list of methods
            self.cw('static struct PyMethodDef ' + t.name + '_methods[] = {{NULL, NULL}};')
            #########################################################################
            # --- And finally, the initialization function
            self.cw('void ' + fname('init' + t.name + 'py') +
                    '(long *i, char *fobj, ForthonObject **cobj__, ' +
                    'long *setinitvalues, long *deallocatable)')
            self.cw('{')
            self.cw('  ForthonObject *obj;')
            self.cw('  obj = (ForthonObject *) PyObject_GC_New(ForthonObject, ' + '&ForthonType);')
            self.cw('  if (*i > 0) {obj->name = ' + pkgname + '_fscalars[*i].name;}')
            self.cw('  else        {obj->name = %spointee__;}'%t.name)
            self.cw('  obj->typename = "' + t.name + '";')
            self.cw('  ' + t.name + 'declarevars(obj);')
            self.cw('  obj->setdims = *' + t.name + 'setdims;')
            self.cw('  obj->setstaticdims = *' + t.name + 'setstaticdims;')
            self.cw('  obj->fmethods = ' + t.name + '_methods;')
            self.cw('  obj->__module__ = Py_BuildValue("s", "%s");'%self.getmodulename())
            self.cw('  obj->fobj = fobj;')
            self.cw('  if (*deallocatable == 1)')
            self.cw('    obj->fobjdeallocate = *' + fname(self.fsub(t, 'deallocatef')) + ';')
            self.cw('  else')
            self.cw('    obj->fobjdeallocate = NULL;')
            self.cw('  obj->nullifycobj = *' + fname(self.fsub(t, 'nullifycobjf')) + ';')
            self.cw('  obj->allocated = 0;')
            self.cw('  obj->garbagecollected = %d;'%garbagecollected)
            self.cw('  *cobj__ = obj;')
            self.cw('  if (PyErr_Occurred()) {')
            self.cw('    PyErr_Print();')
            self.cw('    Py_FatalError("can not initialize type ' + t.name + '");')
            self.cw('    }')
            self.cw('  Forthon_BuildDicts(obj);')
            self.cw('  ForthonPackage_allotdims(obj);')
            self.cw('  ' + fname(self.fsub(t, 'passpointers')) + '(fobj, setinitvalues);')
            self.cw('  ' + fname(self.fsub(t, 'nullifypointers')) + '(fobj);')
            self.cw('  ForthonPackage_staticarrays(obj);')
            if garbagecollected:
                self.cw('  PyObject_GC_Track((PyObject *)obj);')
            self.cw('}')

            #########################################################################
            # --- increments the python reference counter
            # --- Note that if the python object associated with the derived type
            # --- instance had not been created, then do so. In that case, an
            # --- INCREF is not needed since the creation itself effectively
            # --- increments the counter from 0 to 1.
            self.cw('void ' + fname('incref' + t.name + 'py') + '(ForthonObject **cobj__, char *fobj)')
            self.cw('{')
            self.cw('  if (*cobj__ == NULL) {')
            self.cw('    long i=-1, s=0, d=0;')
            self.cw('    ' + fname('init' + t.name + 'py') + '(&i, fobj, cobj__, &s, &d);}')
            self.cw('  else {')
            self.cw('    Py_INCREF(*cobj__);}')
            self.cw('}')
            # --- decrements the python reference counter
            self.cw('void ' + fname('decref' + t.name + 'py') + '(ForthonObject **cobj__)')
            self.cw('{')
            self.cw('  Py_XDECREF(*cobj__);')
            self.cw('}')

            # --- This is called when an object is released by fortran. It is then
            # --- up to any python reference holder to deallocate the object.
            self.cw('void ' + fname(self.fsub(t, 'makedeallocable')) + '(ForthonObject **cobj)')
            self.cw('{')
            self.cw('  (*cobj)->fobjdeallocate = *' + fname(self.fsub(t, 'deallocatef')) + ';')
            self.cw('}')

            #########################################################################
            # --- Write set pointers routine which gets all of the fortran pointers
            self.cw('void ' + fname(self.fsub(t, 'grabscalarpointers')) + '(long *i, char *p, ForthonObject **obj)')
            self.cw('{')
            self.cw('  /* Grabs pointer for the scalar */')
            self.cw('  (*obj)->fscalars[*i].data = (char *)p;')
            self.cw('}')

            self.cw('void ' + fname(self.fsub(t, 'setderivedtypepointers')) + '(long *i, char **p, ForthonObject **obj)')
            self.cw('{')
            self.cw('  /* Grabs pointer for the derived type */')
            self.cw('  (*obj)->fscalars[*i].data = (char *)(*p);')
            self.cw('}')

            self.cw('void ' + fname(self.fsub(t, 'grabarraypointers')) + '(long *i, char *p, ForthonObject **obj)')
            self.cw('{')
            self.cw('  /* Grabs pointer for the array */')
            self.cw('  (*obj)->farrays[*i].data.s = (char *)p;')
            self.cw('}')

            self.cw('void ' + fname(self.fsub(t, 'grabarraypointersobj')) + '(Fortranarray *farray, char *p)')
            self.cw('{')
            self.cw('  /* Grabs pointer for the array */')
            self.cw('  farray->data.s = (char *)p;')
            self.cw('}')

            # --- This routine gets the dimensions from an array. It is called from
            # --- fortran and the last argument should be shape(array).
            # --- This is only used for routines with the fassign attribute.
            # --- I should check the fortran standard - for the dims argument, the
            # --- fortran is passing in shape(x). Apparently, it is the same length
            # --- as longs.
            self.cw('void ' + fname(self.fsub(t, 'setarraydims')) + '(Fortranarray *farray, long *dims)')
            self.cw('{')
            self.cw('  int id;')
            self.cw('  for (id=0;id<farray->nd;id++)')
            self.cw('    farray->dimensions[id] = (npy_intp)(dims[id]);')
            self.cw('}')

            #########################################################################
            #########################################################################

            #########################################################################
            # --- Write out f90 modules, including any data statements
            if writemodules:
                self.fw('')
                g = t.name
                self.fw('module ' + t.name + 'module')

                # --- Check if any variables are derived types. If so, the module
                # --- containing the type must be used. This module though does not
                # --- need to include itself of course.
                printedtypes = [t.name]
                for v in slist + alist:
                    if v.derivedtype:
                        if v.type not in printedtypes:
                            self.fw('  use ' + v.type + 'module')
                            printedtypes.append(v.type)

                self.fw('  save')
                self.fw('  type ' + t.name + '')
                self.fw('    integer(' + isz + '):: cobj__ = 0')
                for s in slist:
                    self.fw('    ' + fvars.ftof(s.type), noreturn=1)
                    if s.dynamic:
                        self.fw(', pointer', noreturn=1)
                    self.fw(':: ' + s.name)
                    # --- data statement is handle by the passpointer routine
                for a in alist:
                    if a.dynamic:
                        if a.type == 'character':
                            self.fw('    character(len=' + a.dims[0].high + '), pointer:: ' +
                                    a.name, noreturn=1)
                            ndims = len(a.dims) - 1
                        else:
                            self.fw('    ' + fvars.ftof(a.type) + ', pointer:: ' + a.name, noreturn=1)
                            ndims = len(a.dims)
                        if ndims > 0:
                            self.fw('(' + (ndims*':,')[:-1] + ')', noreturn=1)
                        self.fw('')
                    else:
                        if a.type == 'character':
                            self.fw('    character(len=' + a.dims[0].high + '):: ' +
                                    a.name + re.sub('[ \t\n]', '', a.dimstring))
                        else:
                            self.fw('    ' + fvars.ftof(a.type) + ':: ' +
                                    a.name + re.sub('[ \t\n]', '', a.dimstring))

                self.fw('  end type ' + t.name + '')

                # --- These functions must be in the module so that its return type
                # --- is defined.
                # --- Note that the body of the New function is replicated below in
                # --- the NewF function. Any changes here should be made there.
                self.fw('contains')
                self.fw('  function New' + t.name + '(deallocatable) result(newobj__)')
                self.fw('    type(' + t.name + '), pointer:: newobj__')
                self.fw('    integer(' + isz + '), optional:: deallocatable')
                self.fw('    integer(' + isz + '):: d')
                self.fw('    integer:: error')
                self.fw('    allocate(newobj__, stat=error)')
                self.fw('    if (error /= 0) then')
                self.fw('      print*, "ERROR during allocation of ' + t.name + '"')
                self.fw('      stop')
                self.fw('    endif')
                for s in slist:
                    if s.dynamic:
                        self.fw('    nullify(newobj__%' + s.name + ')')
                self.fw('    if (present(deallocatable)) then')
                self.fw('      d = deallocatable')
                self.fw('    else')
                self.fw('      d = 0')
                self.fw('    endif')
                self.fw('    call init' + t.name + 'py(int(-1, ' + isz + '), newobj__, newobj__%cobj__, int(1, ' + isz + '), d)')
                self.fw('    return')
                self.fw('  end function New' + t.name + '')
                self.fw('  subroutine Del' + t.name + '(oldobj__)')
                self.fw('    type(' + t.name + '), pointer:: oldobj__')
                self.fw('    call DecRef' + t.name + '(oldobj__)')
                self.fw('    return')
                self.fw('  end subroutine Del' + t.name + '')
                # --- This routine should be called by fortran instead of deallocate.
                # --- The first decref accounts for the reference "owned" by the
                # --- fortran object itself. If there are no further existing python
                # --- references to the object, then it is just deallocated. If there
                # --- are references, it is made deallocatable by python, so that
                # --- when all of those references are deleted, the object can finally
                # --- be deallocated. The fortran pointer to the instance
                # --- is in all cases nullified. This is necessary so that future
                # --- accesses to the object can detect that the pointer is now
                # --- unassociated. Similarly, all pointers to derived types that this
                # --- object refers to are nullified. This isolates the object,
                # --- breaking its connections to other objects. As far as the fortran
                # --- is concerned, this object no longer exists and should not have
                # --- references to other objects.
                self.fw('  subroutine Release' + t.name + '(oldobj__)')
                self.fw('    type(' + t.name + '), pointer:: oldobj__')
                self.fw('    integer:: error')
                self.fw('    if (oldobj__%cobj__ /= 0) then')
                self.fw('      call decref' + t.name + 'py(oldobj__)')
                self.fw('    endif')
                for s in slist:
                    if s.dynamic:
                        self.fw('    if (associated(oldobj__%' + s.name + '))' + 'nullify(oldobj__%' + s.name + ')')
                self.fw('    if (oldobj__%cobj__ == 0) then')
                self.fw('      deallocate(oldobj__, stat=error)')
                self.fw('      if (error /= 0) then')
                self.fw('        print*, "ERROR during deallocation of ' + t.name + '"')
                self.fw('        stop')
                self.fw('      endif')
                self.fw('    else')
                self.fw('      call ' + self.fsub(t, 'makedeallocable') + '(oldobj__%cobj__)')
                self.fw('    endif')
                self.fw('    nullify(oldobj__)')
                self.fw('    return')
                self.fw('  end subroutine Release' + t.name + '')
                # --- This routine is needed to get around limitations
                # --- on subroutine arguments. A variable must be a pointer (or
                # --- allocatable) to be deallocated, but for a pointer or target
                # --- to be passed into a subroutine, its signature must be defined.
                # --- This can't be done from C, so this must be inside a module.
                # --- Furthermore, only a pointer can be passed into a pointer.
                # --- To get around that, the input argument is declared a target,
                # --- then an explicit pointer assignment gives a pointer that can
                # --- be deallocated.
                # --- Note that this fails with the IBM xlf compiler. It takes a
                # --- strict interpretation of the Fortran standard - an error value
                # --- is returned from the deallocate since the pointer points to
                # --- (what appears to be) a static object. So, for now, the
                # --- the error is ignored and the object (probably) left allocated.
                self.fw('  subroutine ' + t.name + 'dealloc(oldobj__)')
                self.fw('    type(' + t.name + '), target:: oldobj__')
                self.fw('    type(' + t.name + '), pointer:: poldobj__')
                self.fw('    integer:: error')
                self.fw('    poldobj__ => oldobj__')
                self.fw('    deallocate(poldobj__, stat=error)')
                if fcompname != 'xlf':
                    self.fw('    if (error /= 0) then')
                    self.fw('      print*, "ERROR during deallocation of ' + t.name + '"')
                    self.fw('      stop')
                    self.fw('    endif')
                self.fw('    return')
                self.fw('  end subroutine ' + t.name + 'dealloc')

                self.fw('end module ' + t.name + 'module')

            # --- These subroutines are written outside of the module in case
            # --- write module is false. This way, they are always created.
            # --- The InitPyRef and DecRef are called by the New and Del routines
            # --- if the modules are written. They are also meant to be explicitly
            # --- called from the users Fortran code if the creation and deletion
            # --- of derived type instances is done there.
            # --- The deallocatable should be false whenever the variable is not
            # --- a pointer or allocatable.  Otherwise the deallocate can result
            # --- in a memory error.
            self.fw('subroutine InitPyRef' + t.name + '(newobj__, setinitvalues, ' + 'deallocatable)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: newobj__')
            self.fw('  integer(' + isz + '), optional:: setinitvalues, deallocatable')
            self.fw('  integer(' + isz + '):: s, d')
            self.fw('  if (present(setinitvalues)) then')
            self.fw('    s = setinitvalues')
            self.fw('  else')
            self.fw('    s = 1')
            self.fw('  endif')
            self.fw('  if (present(deallocatable)) then')
            self.fw('    d = deallocatable')
            self.fw('  else')
            self.fw('    d = 1')
            self.fw('  endif')
            self.fw('  call init' + t.name + 'py(int(-1, ' + isz + '), newobj__, newobj__%cobj__, s, d)')
            self.fw('  return')
            self.fw('end subroutine InitPyRef' + t.name)
            self.fw('subroutine IncRef' + t.name + '(fobj__)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: fobj__')
            self.fw('  call incref' + t.name + 'py(fobj__%cobj__, fobj__)')
            self.fw('  return')
            self.fw('end subroutine IncRef' + t.name)
            self.fw('subroutine DecRef' + t.name + '(oldobj__)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: oldobj__')
            self.fw('  integer(' + isz + '):: cobj__')
            self.fw('  cobj__ = oldobj__%cobj__')
            # --- Note that during the decreftypepy call, this object may become
            # --- deallocated (if there are no other references to it). So, this
            # --- assignment must be done first.
            # self.fw('  oldobj__%cobj__ = 0')
            self.fw('  call decref' + t.name + 'py(cobj__)')
            self.fw('  return')
            self.fw('end subroutine DecRef' + t.name)
            # --- This should be called by fortran to check if the object can be
            # --- explicitly deallocated.
            # --- If there are no existing python references to the object, then
            # --- it can be deallocated. If there are references, it can not
            # --- be deallocated yet. It is made deallocatable by python, so
            # --- that when all of those references are deleted, the object can
            # --- be deallocated.
            self.fw('function ' + t.name + 'deallocatable(oldobj__) result(d)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '), pointer:: oldobj__')
            self.fw('  logical:: d')
            self.fw('  if (oldobj__%cobj__ == 0) then')
            self.fw('    d = .true.')
            self.fw('  else')
            self.fw('    call ' + self.fsub(t, 'makedeallocable') + '(oldobj__%cobj__)')
            self.fw('    d = .false.')
            self.fw('  endif')
            self.fw('  return')
            self.fw('end function ' + t.name + 'deallocatable')

            # --- The memory handling routines check if the cobj__ has been
            # --- set - if not, then set it.
            # --- XXX
            # --- It may not be a good idea to call IncRef if the cobj does not
            # --- already exist. This is a hidden reference that may lead to
            # --- memory leaks. A better way would be to force the coder to
            # --- either explicitly call IncRef or use the New routine to allocate
            # --- the instance (which also creates a python object).
            self.fw('subroutine ' + t.name + 'allot(fobj__)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: fobj__')
            self.fw('  if (fobj__%cobj__ == 0) call IncRef' + t.name + '(fobj__)')
            self.fw('  call tallot(fobj__%cobj__)')
            self.fw('  return')
            self.fw('end subroutine ' + t.name + 'allot')
            self.fw('subroutine ' + t.name + 'change(fobj__)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: fobj__')
            self.fw('  if (fobj__%cobj__ == 0) call IncRef' + t.name + '(fobj__)')
            self.fw('  call tchange(fobj__%cobj__)')
            self.fw('  return')
            self.fw('end subroutine ' + t.name + 'change')
            self.fw('subroutine ' + t.name + 'free(fobj__)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: fobj__')
            self.fw('  if (fobj__%cobj__ == 0) call IncRef' + t.name + '(fobj__)')
            self.fw('  call tfree(fobj__%cobj__)')
            self.fw('  return')
            self.fw('end subroutine ' + t.name + 'free')

            #########################################################################
            self.fw('! ' + self.fsub(t, 'passpointers', dohash=0))
            self.fw('subroutine ' + self.fsub(t, 'passpointers') + '(fobj__, setinitvalues)')

            # --- Write out the Use statements
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: fobj__')
            self.fw('  integer(' + isz + '):: setinitvalues')

            # --- This is for legacy code that directly referenced obj__ in the initial values
            self.fw('  target:: fobj__')
            self.fw('  type(' + t.name + '), pointer:: obj__')
            self.fw('  obj__ => fobj__')

            # --- Write out calls to c routine passing down pointers to scalars
            for i in range(len(slist)):
                s = slist[i]
                if s.dynamic:
                    continue
                if s.derivedtype:
                    # --- This is only called for static instances, so deallocatable is
                    # --- set to false (the last argument).
                    self.fw('  call init' + s.type + 'py(int(-1, ' + isz + '), fobj__%' + s.name +
                            ', fobj__%' + s.name + '%cobj__, setinitvalues, int(0, ' + isz + '))')
                    self.fw('  call ' + self.fsub(t, 'setderivedtypepointers') + '(' +
                            'int(' + repr(i) + ', ' + isz + '), fobj__%' + s.name + '%cobj__, fobj__%cobj__)')
                else:
                    self.fw('  call ' + self.fsub(t, 'grabscalarpointers') + '(' +
                            'int(' + repr(i) + ', ' + isz + '), fobj__%' + s.name + ', fobj__%cobj__)')

            # --- Write out calls to c routine passing down pointers to arrays
            for i in range(len(alist)):
                a = alist[i]
                if not a.dynamic:
                    if not a.derivedtype:
                        # --- This assumes that a scalar is given which is broadcasted
                        # --- to fill the array.
                        self.fw('  call ' + self.fsub(t, 'grabarraypointers') + '(' + 'int(' + repr(i) + ', ' + isz + ')' +
                                ', fobj__%' + a.name + ', fobj__%cobj__)')

            # --- Set the initial values only if the input flag is 1.
            self.fw('  if (setinitvalues == 1) then')
            for s in slist:
                if s.dynamic or s.derivedtype or not s.data:
                    continue
                self.fw('    fobj__%' + s.name + ' = ' + s.data[1:-1])
            for a in alist:
                if a.dynamic or a.derivedtype or not a.data:
                    continue
                self.fw('    fobj__%' + a.name + ' = ' + a.data[1:-1])
            self.fw('  endif')

            # --- Finish the routine
            self.fw('  return')
            self.fw('end')

            #########################################################################
            # --- Nullifies the pointers of all dynamic variables. This is needed
            # --- since in some compilers, the associated routine returns
            # --- erroneous information if the status of a pointer is undefined.
            # --- Pointers must be explicitly nullified in order to get
            # --- associated to return a false value.
            self.fw('! ' + self.fsub(t, 'nullifypointers', dohash=0))
            self.fw('subroutine ' + self.fsub(t, 'nullifypointers') + '(fobj__)')

            # --- Write out the Use statements
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: fobj__')

            for i in range(len(slist)):
                s = slist[i]
                if s.dynamic:
                    self.fw('  nullify(fobj__%' + s.name + ')')
            for i in range(len(alist)):
                a = alist[i]
                if a.dynamic:
                    self.fw('  nullify(fobj__%' + a.name + ')')

            self.fw('  return')
            self.fw('end')

            #########################################################################
            # --- Write routine for each dynamic variable which gets the pointer
            # --- from the wrapper
            for s in slist:
                if (s.dynamic or s.derivedtype) and not s.parameter:
                    self.fw('! ' + self.fsub(t, 'setscalarpointer', s.name, dohash=0))
                    self.fw('subroutine ' + self.fsub(t, 'setscalarpointer', s.name) + '(p__, fobj__, nullit__)')
                    self.fw('  use ' + t.name + 'module')
                    self.fw('  type(' + t.name + '):: fobj__')
                    self.fw('  ' + fvars.ftof(s.type) + ', target:: p__')
                    self.fw('  integer(' + isz + '):: nullit__')
                    if s.dynamic:
                        self.fw('  if (nullit__ == 0) then')
                        self.fw('    fobj__%' + s.name + ' => p__')
                        self.fw('  else')
                        self.fw('    nullify(fobj__%' + s.name + ')')
                        self.fw('  endif')
                    else:
                        self.fw('  fobj__%' + s.name + ' = p__')
                    self.fw('  return')
                    self.fw('end')
                if s.dynamic:
                    self.fw('subroutine ' + self.fsub(t, 'getscalarpointer', s.name) + '(cobj__, fobj__, createnew__)')
                    self.fw('  use ' + t.name + 'module')
                    self.fw('  integer(' + isz + '):: cobj__')
                    self.fw('  integer(4):: createnew__')
                    self.fw('  type(' + t.name + '):: fobj__')
                    self.fw('  if (associated(fobj__%' + s.name + ')) then')
                    self.fw('    if (fobj__%' + s.name + '%cobj__ == 0 .and. createnew__ == 1) then')
                    self.fw('      call init' + s.type + 'py(int(-1, ' + isz + '), fobj__%' + s.name + ', ' +
                            'fobj__%' + s.name + '%cobj__, int(0, ' + isz + '), int(0, ' + isz + '))')
                    self.fw('    endif')
                    self.fw('    cobj__ = fobj__%' + s.name + '%cobj__')
                    self.fw('  else')
                    self.fw('    cobj__ = 0')
                    self.fw('  endif')
                    self.fw('  return')
                    self.fw('end')

            for a in alist:
                if a.dynamic:
                    self.fw('! ' + self.fsub(t, 'setarraypointer', a.name, dohash=0))
                    self.fw('subroutine ' + self.fsub(t, 'setarraypointer', a.name) + '(p__, fobj__, dims__)')
                    self.fw('  use ' + t.name + 'module')
                    self.fw('  type(' + t.name + '):: fobj__')
                    self.fw('  integer(' + isz + '):: dims__(' + repr(len(a.dims)) + ')')
                    if a.type == 'character':
                        self.fw('  character(len=' + a.dims[0].high + '), target:: ' +
                                'p__' + self.prefixdimsf(re.sub('[ \t\n]', '', a.dimstring)) + '')
                    else:
                        self.fw('  ' + fvars.ftof(a.type) + ', target:: ' +
                                'p__' + self.prefixdimsf(re.sub('[ \t\n]', '', a.dimstring)) + '')
                    self.fw('  fobj__%' + a.name + ' => p__')
                    self.fw('  return')
                    self.fw('end')
                    if a.dynamic or re.search('fassign', a.attr):
                        self.fw('! ' + self.fsub(t, 'getarraypointer', a.name, dohash=0))
                        self.fw('subroutine ' + self.fsub(t, 'getarraypointer', a.name) + '(farray__, fobj__)')
                        self.fw('  use ' + t.name + 'module')
                        self.fw('  integer(' + isz + '):: farray__')
                        self.fw('  integer(' + isz + '):: ss(%d)'%(len(a.dims)))
                        self.fw('  type(' + t.name + '):: fobj__')
                        self.fw('  if (.not. associated(fobj__%' + a.name + ')) return ')
                        self.fw('  call ' + self.fsub(t, 'grabarraypointersobj') + '(farray__, fobj__%' + a.name + ')')
                        if a.type == 'character':
                            self.fw('  ss(1:%d)'%(len(a.dims)-1) + ' = shape(fobj__%' + a.name + ')')
                            self.fw('  ss(%d)'%(len(a.dims)) + ' = %s'%a.dims[0].high)
                        else:
                            self.fw('  ss = shape(fobj__%' + a.name + ')')
                        self.fw('  call ' + self.fsub(t, 'setarraydims') + '(farray__, ss)')
                        self.fw('  return')
                        self.fw('end')

            #########################################################################
            # --- Write the routine which creates a new instance of the derived type
            # --- Note that part of the body of this routine is taken from the New
            # --- routine above. Any change in one should be copied to the other.
            # --- The body is copied from New since in cases where the modules
            # --- are not written out, the New routine will not exist.
            self.fw('function ' + self.fsub(t, 'NewF') + '() result(cobj__)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  integer(' + isz + '):: cobj__')
            self.fw('  integer:: error')
            self.fw('  type(' + t.name + '), pointer:: newobj__')
            self.fw('  allocate(newobj__, stat=error)')
            self.fw('  if (error /= 0) then')
            self.fw('    print*, "ERROR during allocation of ' + t.name + '"')
            self.fw('    stop')
            self.fw('  endif')
            for s in slist:
                if s.dynamic:
                    self.fw('  nullify(newobj__%' + s.name + ')')
            # self.fw('  call InitPyRef' + t.name + '(newobj__, 1, 1)')
            self.fw('  call init' + t.name + 'py(int(-1, ' + isz + '), newobj__, newobj__%cobj__, int(1, ' + isz + '), int(1, ' + isz + '))')
            self.fw('  cobj__ = newobj__%cobj__')
            self.fw('  return')
            self.fw('end')

            self.fw('subroutine ' + self.fsub(t, 'deallocatef') + '(oldobj__)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: oldobj__')
            self.fw('  call ' + t.name + 'dealloc(oldobj__)')
            self.fw('  return')
            self.fw('end subroutine ' + self.fsub(t, 'deallocatef'))
            self.fw('subroutine ' + self.fsub(t, 'nullifycobjf') + '(fobj__)')
            self.fw('  use ' + t.name + 'module')
            self.fw('  type(' + t.name + '):: fobj__')
            self.fw('  fobj__%cobj__ = 0')
            self.fw('  return')
            self.fw('end subroutine ' + self.fsub(t, 'nullifycobjf'))

#     self.fw('subroutine ' + self.fsub(t, 'DelF') + '(oldobj__)')
#     self.fw('  use ' + t.name + 'module')
#     self.fw('  type(' + t.name + '), pointer:: oldobj__')
#     self.fw('  integer:: error')
#     self.fw('  call DecRef' + t.name + '(oldobj__)')
#     self.fw('  deallocate(oldobj__, stat=error)')
#     self.fw('  if (error /= 0) then')
#     self.fw('    print*, "ERROR during deallocation of ' + t.name + '"')
#     self.fw('    stop')
#     self.fw('  endif')
#     self.fw('  return')
#     self.fw('end subroutine ' + self.fsub(t, 'DelF'))



# Created by David P. Grote, March 6, 1998

# Routines which allows c functions to be callable by fortran
import sys
import re
import struct

from Forthon_options import args

# Set default values of inputs
machine = args.machine
realsize = args.realsize
underscoring = args.underscoring
twounderscores = args.twounderscores

# ----------------------------------------------------------------------------
# Set size of fortran integers and logicals, which is the same size as a
# C long integer. This is always either 4 or 8.
intsize = '%d'%struct.calcsize('l')
isz = 'kind=%s'%intsize

# ----------------------------------------------------------------------------
# Set the size of floating point numbers
fpsize = 'kind=%s'%realsize

# ----------------------------------------------------------------------------
# Creates a function which converts a C name into a Fortran name

if machine in ['aix4', 'aix5', 'win32', 'MAC']:
    def fname(n):
        return n.lower()
elif machine in ['linux', 'linux2', 'linux3', 'darwin', 'SOL', 'AXP', 'DOS', 'cygwin']:
    if underscoring:
        if twounderscores:
            def fname(n):
                m = re.search('_', n)
                if m is None:
                    return n.lower() + '_'
                else:
                    return n.lower() + '__'
        else:
            def fname(n):
                return n.lower() + '_'
    else:
        def fname(n):
            return n.lower()
else:
    raise ValueError('Machine %s not supported'%machine)

# ----------------------------------------------------------------------------
# Sets up C macros which are used to take the place of the length
# of a string passed from Fortran to C.

if machine in ['linux', 'linux2', 'linux3', 'darwin', 'SOL', 'DOS', 'aix4', 'aix5', 'win32', 'cygwin']:
    charlen_at_end = 1
    forthonf2c = """
#define FSTRING char*
#define SL1 ,int sl1
#define SL2 ,int sl2
#define SL3 ,int sl3
#define SL4 ,int sl4
#define SL5 ,int sl5
#define FSTRLEN1(S) sl1
#define FSTRLEN2(S) sl2
#define FSTRLEN3(S) sl3
#define FSTRLEN4(S) sl4
#define FSTRLEN5(S) sl5
#define FSTRPTR(S) S
#define FSETSTRING(S,P,L) S = P
#define FSETSTRLEN1(S,L) sl1 = L
#define FSETSTRLEN2(S,L) sl2 = L
#define FSETSTRLEN3(S,L) sl3 = L
#define FSETSTRLEN4(S,L) sl4 = L
#define FSETSTRLEN5(S,L) sl5 = L
"""

elif machine in ['MAC']:
    charlen_at_end = 0
    forthonf2c = """
#define FSTRING char*
#define SL1
#define SL2
#define SL3
#define SL4
#define SL5
#define FSTRLEN1(S) sl1
#define FSTRLEN2(S) sl2
#define FSTRLEN3(S) sl3
#define FSTRLEN4(S) sl4
#define FSTRLEN5(S) sl5
#define FSTRPTR(S) S
#define FSETSTRING(S,P,L) S = P
#define FSETSTRLEN1(S,L) sl1 = L
#define FSETSTRLEN2(S,L) sl2 = L
#define FSETSTRLEN3(S,L) sl3 = L
#define FSETSTRLEN4(S,L) sl4 = L
#define FSETSTRLEN5(S,L) sl5 = L
"""

else:
    raise ValueError('Machine %s not supported'%machine)


def writeforthonf2c():
    # --- Create the forthonf2c.h file
    ff = open('forthonf2c.h', 'w')
    ff.write(forthonf2c)
    ff.close()



#!/usr/bin/env python
# Created by David P. Grote, March 6, 1998

from cfinterface import *

# --- This is a bit of a hack. The preprocessor takes many of the same arguments as Forthon, but not all.
# --- It uses Forthon_options to avoid duplication.
# --- It uses the first two positional arguments as the input and output filenames.
from Forthon_options import args

def py_ifelse(m, v, t, f=''):
    if m==v:
        return t
    else:
        return f

def main():
    infile = args.pkgname
    outfile = args.remainder[0]

    with open(infile, 'r') as fin:
        text = fin.readlines()

    with open(outfile, 'w') as fout:

        for line in text:
            if line.startswith('%'):
                fout.write(eval(line[1:], globals()))
            else:
                fout.write(line)

if __name__ == '__main__':
    main()




"""Process and build a Forthon package"""

import sys
import os
import re
import distutils
import distutils.sysconfig
from distutils.core import setup, Extension
from distutils.dist import Distribution
from distutils.command.build import build

from Forthon_options import args
from Forthon.compilers import FCompiler

# --- Get the package name, which is assumed to be the first argument.
pkg = args.pkgname

print "Building package " + pkg

# --- Get any extra fortran, C or object files listed.
# --- This scans through args until the end or until it finds an optional
# --- argument (that begins with a '-'). Any remaining optional arguments
# --- are passed to distutils.
remainder = args.remainder
extrafiles = []
while remainder:
    if not remainder[0].startswith('-'):
        extrafiles.append(remainder[0])
        del remainder[0]
    else:
        break

distutil_args = remainder

# --- Default values for command line args
build_base     = args.build_base
build_temp     = args.build_temp
builddir       = args.builddir
cargs          = args.cargs
compile_first  = args.compile_first
debug          = args.debug
defines        = args.defines
dependencies   = args.dependencies
dobuild        = args.dobuild
f90            = args.f90
fargslist      = args.fargslist
fcomp          = args.fcomp
fcompexec      = args.fcompexec
fixed_suffix   = args.fixed_suffix
fopt           = args.fopt
fortranfile    = args.fortranfile
free_suffix    = args.free_suffix
implicitnone   = args.implicitnone
includedirs    = args.includedirs
initialgallot  = args.initialgallot
interfacefile  = args.interfacefile or (pkg + '.v')
libdirs        = args.libdirs
libs           = args.libs
machine        = args.machine
othermacros    = args.othermacros
pkgbase        = args.pkgbase
pkgdir         = args.pkgdir
pkgsuffix      = args.pkgsuffix
realsize       = args.realsize
timeroutines   = args.timeroutines
scripts        = args.scripts
twounderscores = args.twounderscores
underscoring   = args.underscoring
verbose        = args.verbose
with_feenableexcept = args.with_feenableexcept
writemodules   = args.writemodules

# --- These args require special handling

if initialgallot:
    initialgallot = '-a'
else:
    initialgallot = ''

if f90:
    f90 = '--f90'
else:
    f90 = '--f77'

fargs = ' '.join(fargslist)

if not fortranfile:
    # --- Find the main fortran file, which should have a name like pkg.suffix
    # --- where suffix is one of the free or fixed suffices.
    if os.access(pkg + '.' + args.fixed_suffix, os.F_OK):
        fortranfile = pkg + '.' + args.fixed_suffix
    elif os.access(pkg + '.' + args.free_suffix, os.F_OK):
        fortranfile = pkg + '.' + args.free_suffix
    else:
        raise Exception('Main fortran file can not be found, please specify using the fortranfile option')

# --- Set arguments to Forthon, based on defaults and any inputs.
forthonargs = []
if pkgbase is not None: forthonargs.append('--pkgbase %s'%pkgbase)
if pkgsuffix: forthonargs.append('--pkgsuffix %s'%pkgsuffix)
if underscoring: forthonargs.append('--underscoring')
else:            forthonargs.append('--nounderscoring')
if twounderscores: forthonargs.append('--2underscores')
else:              forthonargs.append('--no2underscores')
if not writemodules: forthonargs.append('--nowritemodules')
if timeroutines: forthonargs.append('--timeroutines')

# --- Get the numpy headers path
import numpy
if numpy.__version__ < '1.1.0':
    # --- Older versions of numpy hacked into distutils, changing things
    # --- in such a way to mangle things like object file names. To avoid
    # --- this, the code from get_include is included here explicitly,
    # --- avoiding the importing of numpy.distutils.
    import os
    if numpy.show_config is None:
        # running from numpy source directory
        d = os.path.join(os.path.dirname(numpy.__file__), 'core', 'include')
    else:
        # using installed numpy core headers
        import numpy.core as core
        d = os.path.join(os.path.dirname(core.__file__), 'include')
    includedirs.append(d)
else:
    includedirs.append(numpy.get_include())

# --- Fix path - needed for Cygwin
def fixpath(path, dos=1):
    if machine == 'win32':
        # --- Cygwin does path mangling, requiring two back slashes
        if(dos):
            p = re.sub(r'\\', r'\\\\', path)
            p = re.sub(r':', r':\\\\', p)
        else:
            p = re.sub(r'\\', r'/', path)
#      p = re.sub(r'C:', r'/c', p)
#      if p[1:2] == ':': p = '/cygdrive/'+string.lower(p[0])+p[2:]
        return p
    else:
        return path

# --- Define the seperator to use in paths.
pathsep = os.sep
if machine == 'win32': pathsep = r'\\'

# --- Find place where packages are placed. This imports one of the
# --- Forthon files and gets the path from that. It uses fvars.py since
# --- that is a small file which doesn't have other dependencies.
import fvars
forthonhome = os.path.dirname(fvars.__file__)
forthonhome = fixpath(forthonhome)
del fvars

# --- Reset sys.argv removing the Forthon options and appending any extra
# --- distutils options remaining in the argument list.
# --- This needs to be done for generating builddir since the distutils
# --- options may affect its value.
sys.argv = ['Forthon', 'build']
if build_base:
    sys.argv += ['--build-base', build_base]
if not dobuild:
    sys.argv += ['install']
sys.argv += distutil_args

# --- Find the location of the build directory. There must be a better way
# --- of doing this.
if builddir is None:
    dummydist = Distribution()
    dummydist.parse_command_line()
    dummybuild = dummydist.get_command_obj('build')
    dummybuild.finalize_options()
    builddir = dummybuild.build_temp
    bb = builddir.split(os.sep)
    upbuilddir = len(bb)*(os.pardir + os.sep)
    del dummydist, dummybuild, bb
else:
    upbuilddir = os.getcwd()

if dobuild:
    # --- Add the build-temp option. This is needed since distutils would otherwise
    # --- put the object files from compiling the pkgnamepy.c file in a temp
    # --- directory relative to the file. build_temp defaults to an empty string,
    # --- so the .o files are put in the same place as the .c files.
    sys.argv += ['--build-temp', build_temp]

# --- Add prefix to interfacefile since it will only be referenced from
# --- the build directory.
interfacefile = fixpath(os.path.join(upbuilddir, interfacefile))

# --- Get path to fortranfile relative to the build directory.
upfortranfile = os.path.join(upbuilddir, fortranfile)

# --- Pick the fortran compiler
fcompiler = FCompiler(machine=machine,
                      debug=debug,
                      fcompname=fcomp,
                      fcompexec=fcompexec,
                      implicitnone=implicitnone,
                      twounderscores=twounderscores)

# --- Create some locals which are needed for strings below.
f90free = fcompiler.f90free
f90fixed = fcompiler.f90fixed
popt = fcompiler.popt
forthonargs = forthonargs + fcompiler.forthonargs
if fopt is None: fopt = fcompiler.fopt
extra_link_args = fcompiler.extra_link_args
extra_compile_args = fcompiler.extra_compile_args
define_macros = fcompiler.define_macros

# --- Create path to fortran files for the Makefile since they will be
# --- referenced from the build directory.
freepath = os.path.join(upbuilddir, '%%.%(free_suffix)s'%locals())
fixedpath = os.path.join(upbuilddir, '%%.%(fixed_suffix)s'%locals())

# --- Find location of the python libraries and executable.
prefix = fixpath(sys.prefix, dos=0)
pyvers = sys.version[:3]
python = fixpath(sys.executable, dos=0)

# --- Generate list of package dependencies
dep = ''
for d in dependencies:
    dep = dep + ' -d %s.scalars'%d

sourcedirs = []
def getpathbasename(f):
    dirname, basename = os.path.split(f)
    rootname, suffix = os.path.splitext(basename)
    if dirname not in sourcedirs:
        sourcedirs.append(dirname)
    return rootname, suffix

# --- Loop over extrafiles. For each fortran file, append the object name
# --- to be used in the makefile and for setup. For each C file, add to a
# --- list to be included in setup. For each object file, add to the list
# --- of extra objects passed to setup. Make a string of the extra files,
# --- with paths relative to the build directory.
# --- Also, for fortran files, keep a list of suffices so that the appropriate
# --- build rules can be added to the makefile.
extrafilesstr = ''
extraobjectsstr = ''
extraobjectslist = []
extracfiles = []
fortransuffices = [fixed_suffix, free_suffix]
if machine == 'win32':
    osuffix = '.obj'
else:
    osuffix = '.o'
for f in extrafiles:
    extrafilesstr = extrafilesstr + ' ' + os.path.join(upbuilddir, f)
    root, suffix = getpathbasename(f)
    if suffix[1:] in ['o', 'obj']:
        extraobjectsstr = extraobjectsstr + ' ' + root + osuffix
        extraobjectslist = extraobjectslist + [root + osuffix]
    elif suffix[1:] in ['F', 'F90', 'f', 'f90', 'for', fixed_suffix, free_suffix]:
        extraobjectsstr = extraobjectsstr + ' ' + root + osuffix
        extraobjectslist = extraobjectslist + [root + osuffix]
        if suffix[1:] not in fortransuffices:
            fortransuffices.append(suffix[1:])
    elif suffix[1:] in ['c', 'cc', 'cpp', 'cxx']:
        extracfiles.append(f)

if compile_first != '':
    compile_firstroot, compile_firstsuffix = getpathbasename(compile_first)
    compile_firstobject = compile_firstroot + osuffix
else:
    compile_firstobject = ''

# --- Make string containing other macros files
othermacstr = ''
for f in othermacros:
    othermacstr = othermacstr + ' --macros ' + os.path.join(upbuilddir, f)

# --- Put any defines in a string that will appear at the beginning of the
# --- makefile.
definesstr = ''
for d in (defines + fcompiler.defines):
    definesstr = definesstr + d + '\n'

# --- Define default rule.
fortranroot, fortransuffix = getpathbasename(fortranfile)
defaultrule = 'dependencies: %(compile_firstobject)s %(pkg)s_p%(osuffix)s %(fortranroot)s%(osuffix)s %(pkg)spymodule.c Forthon.h Forthon.c %(extraobjectsstr)s'%locals()

if writemodules:
    # --- Fortran modules are written by the wrapper to the _p file.
    # --- The rest of the fortran files will depend in this ones obhect file.
    # --- This file doesn't depend on any fortran files.
    modulecontainer = pkg+'_p'
    wrapperdependency = ''
else:
    # --- If nowritemodules is set, then it is assumed that modules are contained
    # --- in the main fortran file. In this case, set the dependencies in the
    # --- makefile so that all files depend on the main file, rather than the
    # --- wrapper fortran file.
    modulecontainer = fortranroot
    wrapperdependency = fortranroot+osuffix

# --- convert list of fortranargs into a string
forthonargs = ' '.join(forthonargs)

# --- Add any includedirs to fargs
for i in includedirs:
    fargs = fargs + ' -I'+i+' '

# --- Add in any user supplied cargs
if cargs is not None:
    extra_compile_args.append(cargs)

pypreproc = '%(python)s -c "from Forthon.preprocess import main;main()" %(f90)s -t%(machine)s %(forthonargs)s'%locals()
forthon = '%(python)s -c "from Forthon.wrappergenerator import wrappergenerator_main;wrappergenerator_main()"'%locals()
noprintdirectory = ''
if not verbose:
    # --- Set so that the make doesn't produce any output
    f90fixed = "@echo ' ' F90Fixed $(<F);" + f90fixed
    f90free = "@echo ' ' F90Free $(<F);" + f90free
    pypreproc = "@echo ' ' Preprocess $(<F);" + pypreproc
    forthon = "@echo ' ' Forthon $(<F);" + forthon
    noprintdirectory = '--no-print-directory'

# --- Create a separate rule to compile the compiler_first file, setting it up
# --- so that it doesn't have any dependencies beyond itself.
compile_firstrule = ''
if compile_first != '':
    compile_firstrule = '%(pkg)s_p%(osuffix)s %(fortranroot)s%(osuffix)s %(extraobjectsstr)s: %(compile_firstobject)s\n'%locals()
    if compile_firstsuffix != '':
        suffixpath = os.path.join(upbuilddir, '%(compile_first)s'%locals())
        if compile_firstsuffix[-2:] == '90': ff = f90free
        else:                                ff = f90fixed
        compile_firstrule = """
%(compile_firstobject)s: %(suffixpath)s
	%(ff)s %(fopt)s %(fargs)s -c $<
  """%locals()

# --- Add build rules for fortran files with suffices other than the
# --- basic fixed and free ones. Those first two suffices are included
# --- explicitly in the makefile. Note that this depends on fargs.
extrafortranrules = ''
if len(fortransuffices) > 2:
    for suffix in fortransuffices[2:]:
        suffixpath = os.path.join(upbuilddir, '%%.%(suffix)s'%locals())
        if suffix[-2:] == '90': ff = f90free
        else:                   ff = f90fixed
        extrafortranrules += """
%%%(osuffix)s: %(suffixpath)s %(modulecontainer)s%(osuffix)s
	%(ff)s %(fopt)s %(fargs)s -c $<
    """%locals()
        del suffix, suffixpath, ff

compilerulestemplate = """
%%%(osuffix)s: %(fixedpath)s %(modulecontainer)s%(osuffix)s
	%(f90fixed)s %(fopt)s %(fargs)s -c $<
%%%(osuffix)s: %(freepath)s %(modulecontainer)s%(osuffix)s
	%(f90free)s %(fopt)s %(fargs)s -c $<
"""

if not writemodules and not fortranfile == compile_first:
    # --- If not writing modules, create a rule to compile the main fortran file
    # --- that does not have a dependency on itself. That self dependency causes
    # --- problems when the fortran file is in a subdirectory.
    # --- Skip this if the fortranfile is the same as compile_first to avoid the
    # --- redundant dependency in the makefile.
    if fortransuffix[-2:] == '90': ff = f90free
    else:                          ff = f90fixed
    compilerules = """
%(wrapperdependency)s: %(upfortranfile)s
	%(ff)s %(fopt)s %(fargs)s -c $<
"""%locals()
else:
    compilerules = ''

for sourcedir in sourcedirs:
    # --- Create path to fortran files for the Makefile since they will be
    # --- referenced from the build directory.
    freepath = os.path.join(os.path.join(upbuilddir, sourcedir), '%%.%(free_suffix)s'%locals())
    fixedpath = os.path.join(os.path.join(upbuilddir, sourcedir), '%%.%(fixed_suffix)s'%locals())
    compilerules += compilerulestemplate%locals()

# --- First, create Makefile.pkg which has all the needed definitions
# --- Note the two rules for the pymodule file. The first specifies that the
# --- pymodule.c file should be recreated if the .v file was updatd.
# --- The second changes the timestamp of the pymodule.c to force a rebuild
# --- of the .so during the distutils setup if any of the source files have
# --- been updated.
makefiletext = """
%(definesstr)s

%(defaultrule)s

%(compile_firstrule)s
%(compilerules)s
%(extrafortranrules)s
Forthon.h:%(forthonhome)s%(pathsep)sForthon.h
	%(pypreproc)s %(forthonhome)s%(pathsep)sForthon.h Forthon.h
Forthon.c:%(forthonhome)s%(pathsep)sForthon.c
	%(pypreproc)s %(forthonhome)s%(pathsep)sForthon.c Forthon.c

%(pkg)s_p%(osuffix)s:%(pkg)s_p.%(free_suffix)s %(wrapperdependency)s
	%(f90free)s %(popt)s %(fargs)s -c %(pkg)s_p.%(free_suffix)s
%(pkg)s_p.%(free_suffix)s: %(pkg)spymodule.c
%(pkg)spymodule.c::%(interfacefile)s
	%(forthon)s --realsize %(realsize)s %(f90)s -t %(machine)s %(forthonargs)s %(initialgallot)s %(othermacstr)s %(dep)s %(pkg)s %(interfacefile)s
%(pkg)spymodule.c:: %(upfortranfile)s %(extrafilesstr)s
	@touch %(pkg)spymodule.c
clean:
	rm -rf *%(osuffix)s *_p.%(free_suffix)s *.mod *module.c *.scalars *.so Forthon.c Forthon.h forthonf2c.h build
"""%(locals())
builddir=fixpath(builddir, 0)
try: os.makedirs(builddir)
except: pass
makefile = open(os.path.join(builddir, 'Makefile.%s'%pkg), 'w')
makefile.write(makefiletext)
makefile.close()

# --- Now, execuate the make command.
os.chdir(builddir)
m = os.system('make -f Makefile.%(pkg)s %(noprintdirectory)s'%locals())
if m != 0:
    # --- If there was a problem with the make, then quite this too.
    # --- The factor of 256 just selects out the higher of the two bytes
    # --- returned by system. The upper has the error number returned by make.
    sys.exit(int(m/256))
os.chdir(upbuilddir)

# --- Make sure that the shared object is deleted. This is needed since
# --- distutils doesn't seem to check if objects passed in are newer
# --- than the shared object. The 'try' is used in case the file doesn't
# --- exist (like when the code is built the first time).
try:
    os.remove(pkg+pkgsuffix+'py.so')
except:
    pass

cfiles = [os.path.join(builddir, p) for p in [pkg+'pymodule.c', 'Forthon.c']]
ofiles = [os.path.join(builddir, p) for p in [fortranroot+osuffix,
                                             pkg+'_p'+osuffix] +
                                             extraobjectslist]

# --- DOS requires an extra argument and include directory to build properly
if machine == 'win32': sys.argv.append('--compiler=mingw32')
if machine == 'win32': includedirs+=['/usr/include']

# --- On darwin machines, the python makefile mucks up the -arch argument.
# --- This fixes it.
if machine == 'darwin':
# --- Machines running csh/tcsh seem to have MACHTYPE defined and this is the safest way to set -arch.
    if 'MACHTYPE' in os.environ:
        if os.environ['MACHTYPE'] == 'i386':
            os.environ['ARCHFLAGS'] = '-arch i386'
        elif os.environ['MACHTYPE'] == 'x86_64':
            os.environ['ARCHFLAGS'] = '-arch x86_64'
        elif os.environ['MACHTYPE'] == 'powerpc':
            os.environ['ARCHFLAGS'] = '-arch ppc'
#---  If the shell is bash, MACHTYPE is undefined.  So get what we can from uname. We will assume that if
#---  we are running Snow Leopard we are -arch x86-64 and if running Leopard on intel we are -arch i386.
#---  This can be over-ridden by defining MACHTYPE.
    else:
        archtype = os.uname()[-1]
        if archtype in ['Power Macintosh', 'ppc']:
            os.environ['ARCHFLAGS'] = '-arch ppc'
        elif archtype in ['i386', 'x86_64']:
            kernel_major = eval(os.uname()[2].split('.')[0])
            if kernel_major < 10 :
                os.environ['ARCHFLAGS'] = '-arch i386'  # Leopard or earlier
            else:
                os.environ['ARCHFLAGS'] = '-arch x86_64'  # Snow Leopard

if not verbose:
    print "  Setup " + pkg
    sys.stdout = open(os.devnull, 'w')

if with_feenableexcept:
    define_macros.append(('WITH_FEENABLEEXCEPT', 1))

if pkgbase is None:
    pkgbase = pkg + pkgsuffix

define_macros.append(('FORTHON_PKGNAME', '"%s"'%pkgbase))

package_dir = None
packages = None
if not dobuild:
    # --- When installing, add the package directory if specified so that the
    # --- other files are installed.
    if pkgdir is not None:
        package_dir = {pkgbase:pkgdir}
        packages = [pkgbase]

setup(name = pkgbase,
      packages = packages,
      package_dir = package_dir,
      ext_modules = [Extension('.'.join([pkgbase, pkg+pkgsuffix+'py']),
                               cfiles+extracfiles,
                               include_dirs=[forthonhome]+includedirs,
                               extra_objects=ofiles,
                               library_dirs=fcompiler.libdirs+libdirs,
                               libraries=fcompiler.libs+libs,
                               define_macros=define_macros,
                               extra_compile_args=extra_compile_args,
                               extra_link_args=extra_link_args)],
      scripts = scripts
     )

